# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
Abstract tokenizer that can be extended to be specialized in different languages.

This tokenizer is already initialized to read integers, floats, strings and basic punctuation.
"""

from docplex.cp.utils import to_internal_string, is_string, StringIO, TextFileLineReader
import os


###############################################################################
## Utility classes
###############################################################################

class Token(object):
    """ Token returned by tokenizer  """
    __slots__ = ('type',   # Token type
                 'value',  # Token string value (with quotes for strings)
                 )

    def __init__(self, type, value):
        """ Create a new token
        Args:
            type:  Token type
            value: Token value
        """
        super(Token, self).__init__()
        self.type = type
        self.value = value

    def get_string(self):
        """ Get the string corresponding to the value, interpreting escape sequences if necessary
        Returns:
            Expanded string value
        """
        return self.value if self.type != TOKEN_TYPE_STRING else to_internal_string(self.value[1:-1])

    def __str__(self):
        """ Build a string representing this token
        Returns:
            String representing this token
        """
        return self.value

    def __eq__(self, other):
        """ Check if this token is equal to another object
        Args:
            other:  Object to compare with
        Returns:
            True if 'other' is a token with the same value then this one, False otherwise
        """
        return (other is self) or (isinstance(other, Token) and (other.type == self.type) and (other.value == self.value))

    def __ne__(self, other):
        """ Check if this token is different than another object
        Args:
            other:  Object to compare with
        Returns:
            True if 'other' is not a token or a different token than this one, False otherwise
        """
        return not self.__eq__(other)


###############################################################################
## Constants
###############################################################################

# Token types
TOKEN_TYPE_NONE        = "None"         # No token (for EOF)
TOKEN_TYPE_INTEGER     = "Integer"      # Integer number
TOKEN_TYPE_FLOAT       = "Float"        # Floating point number
TOKEN_TYPE_PUNCTUATION = "Punctuation"  # Ponctuation, like comma, colon, semicolon, brackets, etc
TOKEN_TYPE_OPERATOR    = "Operator"     # Operator
TOKEN_TYPE_KEYWORD     = "Keyword"      # Keyword (reserved symbol)
TOKEN_TYPE_SYMBOL      = "Symbol"       # Symbol (identifier)
TOKEN_TYPE_STRING      = "String"       # String inside double quotes, with Java/C++/Python escape characters
TOKEN_TYPE_VERSION     = "Version"      # Version number expressed as integers separated by dots
TOKEN_TYPE_COMMENT     = "Comment"      # Comment

# Predefined tokens
TOKEN_EOF = Token(TOKEN_TYPE_NONE, "EOF")

TOKEN_ASSIGN         = Token(TOKEN_TYPE_OPERATOR, "=")
TOKEN_EQUAL          = Token(TOKEN_TYPE_OPERATOR, "==")
TOKEN_DIFFERENT      = Token(TOKEN_TYPE_OPERATOR, "!=")
TOKEN_GREATER        = Token(TOKEN_TYPE_OPERATOR, ">")
TOKEN_GREATER_EQUAL  = Token(TOKEN_TYPE_OPERATOR, ">=")
TOKEN_LOWER          = Token(TOKEN_TYPE_OPERATOR, "<")
TOKEN_LOWER_EQUAL    = Token(TOKEN_TYPE_OPERATOR, "<=")

TOKEN_PLUS           = Token(TOKEN_TYPE_OPERATOR, "+")
TOKEN_MINUS          = Token(TOKEN_TYPE_OPERATOR, "-")
TOKEN_STAR           = Token(TOKEN_TYPE_OPERATOR, "*")
TOKEN_SLASH          = Token(TOKEN_TYPE_OPERATOR, "/")
TOKEN_PERCENT        = Token(TOKEN_TYPE_OPERATOR, "%")
TOKEN_CIRCUMFLEX     = Token(TOKEN_TYPE_OPERATOR, "^")

TOKEN_DOT            = Token(TOKEN_TYPE_PUNCTUATION, ".")
TOKEN_COMMA          = Token(TOKEN_TYPE_PUNCTUATION, ",")
TOKEN_COLON          = Token(TOKEN_TYPE_PUNCTUATION, ":")
TOKEN_SEMICOLON      = Token(TOKEN_TYPE_PUNCTUATION, ";")
TOKEN_HOOK_OPEN      = Token(TOKEN_TYPE_PUNCTUATION, "[")
TOKEN_HOOK_CLOSE     = Token(TOKEN_TYPE_PUNCTUATION, "]")
TOKEN_BRACE_OPEN     = Token(TOKEN_TYPE_PUNCTUATION, "{")
TOKEN_BRACE_CLOSE    = Token(TOKEN_TYPE_PUNCTUATION, "}")
TOKEN_PARENT_OPEN    = Token(TOKEN_TYPE_PUNCTUATION, "(")
TOKEN_PARENT_CLOSE   = Token(TOKEN_TYPE_PUNCTUATION, ")")

TOKEN_PIPE           = Token(TOKEN_TYPE_PUNCTUATION, "|")
TOKEN_HASH           = Token(TOKEN_TYPE_PUNCTUATION, '#')
TOKEN_BANG           = Token(TOKEN_TYPE_PUNCTUATION, '!')

TOKEN_KEYWORD_TRUE   = Token(TOKEN_TYPE_KEYWORD, "true")
TOKEN_KEYWORD_FALSE  = Token(TOKEN_TYPE_KEYWORD, "false")

_SYMBOL_START_CHARS = set("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_")
_SYMBOL_BODY_CHARS  = set("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789")

###############################################################################
## Public classes
###############################################################################

class Tokenizer(object):
    """ Tokenizer for CPO file format """
    __slots__ = ('name',                 # Input name (for error string build)
                 'input',                # Input stream
                 'line',                 # Current input line
                 'current_char',         # Last read character
                 'token_start',          # Index of token start in current line
                 'line_length',          # Current input line length
                 'read_index',           # Current read index in the line
                 'line_number',          # Current line number
                 'skip_comments',        # Skip comments indicator
                 'symbols',              # Map of symbols. Key is name, value is symbol token
                 'char_handlers',        # Map of functions processing characters
                 'default_char_handler', # Default character handler
                 'first_in_line',        # First token in line indicator
                 )

    def __init__(self, name=None, input=None, file=None, encoding='utf-8-sig'):
        """ Create a new tokenizer
        Args:
            name:     (Optional) Name of the stream
            input:    (Optional) Input stream or string
            file:     (Optional) Input file
            encoding: (Optional) Character encoding, utf-8 by default
        """
        super(Tokenizer, self).__init__()

        # Open file if any
        if file is not None:
            self.input = TextFileLineReader(file, encoding)
        # Or get input if any
        elif input is not None:
            if is_string(input):
                self.input = StringIO(input)
            else:
                self.input = input
        # Set name
        self.name = name
        if name is None:
            self.name = file

        # Initialize reading
        self.line = ""
        self.current_char = ''
        self.line_length = 0
        self.read_index = 0
        self.line_number = 0
        self.token_start = 0
        self.skip_comments = True
        self.default_char_handler = None

        # Initialize symbols table (allows use of 'is' instead of '==')
        self.symbols = {}

        # Initialize array of character processors
        self.char_handlers = [None] * 128
        for c in _SYMBOL_START_CHARS:
            self.add_char_handler(c, self._read_symbol)
        for c in "0123456789":
            self.add_char_handler(c, self._read_number)
        self.add_char_handler('"', self._read_string)
        self.add_char_handler('>', self._read_greater)
        self.add_char_handler('<', self._read_lower)
        self.add_char_handler('=', self._read_equal)
        self.add_char_handler('!', self._read_bang)

        self.add_char_handler('+', lambda: TOKEN_PLUS)
        self.add_char_handler('-', lambda: TOKEN_MINUS)
        self.add_char_handler('*', lambda: TOKEN_STAR)
        self.add_char_handler('/', lambda: TOKEN_SLASH)
        self.add_char_handler('%', lambda: TOKEN_PERCENT)
        self.add_char_handler('^', lambda: TOKEN_CIRCUMFLEX)
        self.add_char_handler(',', lambda: TOKEN_COMMA)
        self.add_char_handler('.', lambda: TOKEN_DOT)
        self.add_char_handler(':', lambda: TOKEN_COLON)
        self.add_char_handler(';', lambda: TOKEN_SEMICOLON)
        self.add_char_handler('(', lambda: TOKEN_PARENT_OPEN)
        self.add_char_handler(')', lambda: TOKEN_PARENT_CLOSE)
        self.add_char_handler('[', lambda: TOKEN_HOOK_OPEN)
        self.add_char_handler(']', lambda: TOKEN_HOOK_CLOSE)
        self.add_char_handler('{', lambda: TOKEN_BRACE_OPEN)
        self.add_char_handler('}', lambda: TOKEN_BRACE_CLOSE)
        self.add_char_handler('|', lambda: TOKEN_PIPE)
        self.add_char_handler('#', lambda: TOKEN_HASH)


    def add_predefined_symbols(self, ltoks):
        """ Add a list of predefined symbols.

        Tokenizer ensure that if the symbol is found, always the same token is returned.
        This allows to check a token using 'is' instead of '=='.

        Args:
            ltoks:  List of symbol tokens
        """
        for t in ltoks:
            self.symbols[t.value] = t


    def clear_all_char_handlers(self):
        """ Clear all character handlers.
        """
        self.char_handlers = [None] * 127


    def add_char_handler(self, c, chdl):
        """ Add a character handler.

        Character handlers are used to properly process a character read from the input stream.
        By default, this tokenizer process symbols, strings, integer, floats, version, and basic
        punctuation tokens that are defined.
        Character handlers should be added or overwritten to handle properly specific tokens, including comments.

        A character handler takes this tokenizer as parameters (self) and returns the read token.

        Args:
            c:     Character starting the token
            chdl:  Character handler
        """
        self.char_handlers[ord(c)] = chdl


    def set_default_char_handler(self, c, chdl):
        """ Set the default character handler.

        Default character handler is called if no character handler is found. By default, it raises an exception.

        Args:
            chdl:  Character handler
        """
        self.default_char_handler = chdl


    def set_skip_comment(self, skc):
        """ Set the skip comment indicator

        When skip comment is True (default), comment are ignored.
        When skip comment is False, comments are returned as TOKEN_COMMENT tokens.

        Args:
            skc:  Skip comment indicator
        """
        self.skip_comments = skc


    def next_token(self):
        """ Get the next token

        Returns:
            Next available token, TOKEN_EOF if end of input
        """
        # Set first in line indicator
        self.first_in_line = self.read_index == 0

        # Read token loop
        skc = self.skip_comments
        while True:
            # Skip separators
            while True:
                c = self._next_char()
                if c is None:
                    return TOKEN_EOF
                if c > ' ':
                    break

            # Reset current token
            self.token_start = self.read_index - 1

            # Retrieve character handler
            oc = ord(c)
            if oc < 128:
                chld = self.char_handlers[oc]
                if chld is None:
                    chld = self.default_char_handler
            else:
                chld = self.default_char_handler
            if chld is None:
                raise SyntaxError(self.build_error_string("No possible token starting by character '{}'".format(c)))

            # Call character handler
            try:
                tk = chld()
                if not skc or tk.type != TOKEN_TYPE_COMMENT:
                    return tk
            except Exception as err:
                raise SyntaxError(self.build_error_string("Error parsing token starting by '{}': {}".format(c, err)))


    def close(self):
        """ Close this tokenizer
        """
        self.input.close()
        self.line = None
        self.current_char = None


    def _read_symbol(self):
        """ Read and return next symbol """
        # Read symbol
        c = self._peek_char()
        while c and (c in _SYMBOL_BODY_CHARS):
            c = self._skip_and_peek_char()
        s = self._get_token()
        t = self.symbols.get(s)
        if not t:
            t = Token(TOKEN_TYPE_SYMBOL, s)
            self.symbols[s] = t
        return t


    def _read_number(self):
        """ Read and return next number """
        # Read number
        typ = TOKEN_TYPE_INTEGER
        c = self._peek_char()
        while c and (c >= '0') and (c <= '9'):
            c = self._skip_and_peek_char()
        if c == '.':
            c = self._skip_and_peek_char()
            if c == '.':
                # Case of '..' used to specify intervals
                self.read_index -= 1
                return Token(typ, self._get_token())
            typ = TOKEN_TYPE_FLOAT
            while c and (c >= '0') and (c <= '9'):
                c = self._skip_and_peek_char()
            if (c == 'e') or (c == 'E'):
                c = self._skip_and_peek_char()
                if (c == '-') or (c == '+'):
                    c = self._skip_and_peek_char()
                while c and (c >= '0') and (c <= '9'):
                    c = self._skip_and_peek_char()
            elif c == '.':
                typ = TOKEN_TYPE_VERSION
                while (c == '.') or ((c >= '0') and (c <= '9')):
                    c = self._skip_and_peek_char()
        elif (c == 'e') or (c == 'E'):
            typ = TOKEN_TYPE_FLOAT
            c = self._skip_and_peek_char()
            if (c == '-') or (c == '+'):
                c = self._skip_and_peek_char()
            while c and (c >= '0') and (c <= '9'):
                c = self._skip_and_peek_char()
        return Token(typ, self._get_token())


    def _read_string(self):
        """ Read and return next string """
        # Read character sequence
        res = ['"']
        c = ''
        while (c is not None) and (c != '"'):
            c = self._next_char()
            res.append(c)
            if c == '\\':
                res.append(self._next_char())
                c = ''
        if c is None:
            raise SyntaxError(self.build_error_string("String not ended before end of stream"))
        return Token(TOKEN_TYPE_STRING, ''.join(res))


    def _read_greater(self):
        """ Read token starting by > """
        if self._peek_char() == '=':
            self._skip_char()
            return TOKEN_GREATER_EQUAL
        return TOKEN_GREATER


    def _read_lower(self):
        """ Read token starting by < """
        if self._peek_char() == '=':
            self._skip_char()
            return TOKEN_LOWER_EQUAL
        return TOKEN_LOWER


    def _read_equal(self):
        """ Read token starting by = """
        if self._peek_char() == '=':
            self._skip_char()
            return TOKEN_EQUAL
        return TOKEN_ASSIGN


    def _read_bang(self):
        """ Read token starting by ! """
        if self._peek_char() == '=':
            self._skip_char()
            return TOKEN_DIFFERENT
        return TOKEN_BANG


    def _get_line_reminder(self):
        """ Get reminder of the line
        Returns:
            Line remainder content, without ending \n
        """
        start = self.read_index
        c = self._next_char()
        while c and (c != '\n'):
            c = self._next_char()
        return self.line[start:self.read_index-1]


    def _skip_to_end_of_line(self):
        """ Skip read index to end of line
        """
        self.read_index = self.line_length


    def _get_token(self):
        """ Get the last read token
        """
        return self.line[self.token_start:self.read_index]


    def _peek_char(self):
        """ Peek (but not get) next input character
        Returns:
            Next available character, None if end of input
        """
        # Check end of stream
        if not self.line:
            return None

        # Return current character
        return self.line[self.read_index] if self.read_index < self.line_length else '\n'


    def _skip_char(self):
        """ Skip next input character
        """
        self.read_index += 1


    def _skip_and_peek_char(self):
        """ Skip next input character and peek next one on the same line
        Returns:
            Next available character, \n if end of line
        """
        self.read_index += 1
        return self.line[self.read_index] if self.read_index < self.line_length else '\n'


    def _next_char(self):
        """ Get next input character
        This function sets the variable current_char with the returned character
        Returns:
            Next available character, None if end of input
        """
        # Check end of stream
        line = self.line
        if line is None:
            return None

        # Check end of line
        if self.read_index >= self.line_length:
            # Read next line and check end of file
            line = self.input.readline()
            self.line_number += 1
            self.line = line
            self.line_length = len(line)
            self.read_index = 0
            self.first_in_line = True
            # Check end of input
            if line == '':
                self.close()
                return None
        c = line[self.read_index]
        self.read_index += 1
        self.current_char = c
        return c


    def _back_char(self):
        """ Go back to previous character """
        if self.read_index > 0:
            self.read_index -= 1


    def _current_char(self):
        """ Get last read character
        Returns:
            Last read character, None if end of input
        """
        return self.current_char


    def build_error_string(self, msg):
        """ Build error string for exception
        """
        # return "Error in '{}' at line {} (\"{}\") index {}: {}".format(self.name, self.line_number, self.rline, self.read_index, msg)
        # Insert error token in current line
        rline = self.line[:self.read_index] + " ### " + self.line[self.read_index:].rstrip()
        ermsg = "Error in '{}' at line {}:{} : {} (\"{}\")".format(self.name, self.line_number, self.read_index, msg, rline)
        return ermsg


    def __iter__(self):
        """  Define tokenizer as an iterator """
        return self


    def next(self):
        """ For iteration of tokens, get the next available token.

        Returns:
            Next token, StopIteration if EOF
        """
        tok = self.next_token()
        if tok is TOKEN_EOF:
            raise StopIteration()
        return tok


    def __next__(self):
        """ Get the next available token (same as next() for compatibility with Python 3)

        Returns:
            Next token
        """
        return self.next()





# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2019
# --------------------------------------------------------------------------

# gendoc: ignore

# Reference:
# https://www.ibm.com/support/knowledgecenter/SSSA5P_12.9.0/ilog.odms.cplex.help/CPLEX/FileFormats/topics/MST.html

# noinspection PyPep8Naming
import xml.etree.ElementTree as xml_elt


from docplex.mp.constants import EffortLevel


class MSTReader(object):
    cplex_solution_tag  = "CPLEXSolution"
    cplex_solutions_tag = "CPLEXSolutions"

    @classmethod
    def to_float(cls, s):
        try:
            return float(s)
        except TypeError:
            return None

    @classmethod
    def to_int(cls, s, fallback=None):
        try:
            return int(s)
        except TypeError:
            return fallback

    @classmethod
    def new_empty_mipstart(cls, mdl, mipstart_name=None):
        return mdl.new_solution(name=mipstart_name)

    @classmethod
    def read_one_solution(cls, xml_solution, mdl, mst_path):
        new_mipstart_sol = cls.new_empty_mipstart(mdl)
        nb_missed = 0
        effort_level = EffortLevel.Auto
        for child in xml_solution:
            if child.tag == "header":
                obj_s = child.attrib.get('objectiveValue')
                if obj_s is not None:
                    new_mipstart_sol.set_objective_value(float(obj_s))
                # an optional mip start name
                sol_name = child.attrib.get('solutionName')
                if sol_name:
                    new_mipstart_sol.set_name(sol_name)
                # effort level: 0,1,2,...
                effort_s = child.attrib.get('MIPStartEffortLevel')
                if effort_s is not None:
                    # expect an integer attribute, then convert it to EffortLevel
                    # default is Auto
                    effort_level = EffortLevel.parse(cls.to_int(effort_s))

            elif child.tag == "variables":
                for v, var_elt in enumerate(child, start=1):
                    var_attribs = var_elt.attrib
                    var_name = var_attribs.get('name')
                    var_value_s = var_attribs.get('value')

                    # check the <variable> taag
                    valid = True

                    # we need a value
                    var_value = cls.to_float(var_value_s)
                    if var_value is None:
                        mdl.warning("Variable element has no float value, {0} was found - pos={1}", var_value_s, v)
                        valid = False

                    var_index = cls.to_int(var_attribs.get('index'))
                    if var_index is None:
                        mdl.warning("Variable element has no index tag - pos={0}", v)
                        valid = False
                    elif var_index < 0:
                        mdl.warning("Variable element has invalid index: {1} - pos={0}", v, var_index)
                        valid = False
                    if not valid:
                        nb_missed += 1
                    else:
                        # look for a variable from name, index
                        dv = (var_name and mdl.get_var_by_name(var_name)) or mdl.get_var_by_index(var_index)

                        if dv is None:
                            mdl.warning('Cannot find matching variable: name={0}, index={1} - value={2}',
                                        var_name, var_index, var_value)
                            nb_missed += 1
                        else:
                            new_mipstart_sol._set_var_value(dv, var_value)
        if nb_missed:
            mdl.warning("Found {0} unmatched variable(s) in file {1}", nb_missed, mst_path)
        # return a tiple (solution, effort)
        if not new_mipstart_sol.number_of_var_values:
            mdl.warning("No variable read from MIP start file {0}", mst_path)
            return None
        else:
            return new_mipstart_sol, effort_level

    @classmethod
    def read_many_solutions(cls, xml_solutions, mdl, mst_path):
        mipstarts = []
        for child in xml_solutions:
            if child.tag == "CPLEXSolution":
                mip_start = cls.read_one_solution(child, mdl, mst_path)
                if mip_start is not None:
                    # expecting a tuple (solution, effort:int)
                    mipstarts.append(mip_start)
            else:
                print("Unexpected tag name: {0} ignored, expecting CPLEXSolution".format(child.tag))

        return mipstarts

    @classmethod
    def read_root(cls, root, mdl, mst_path):
        root_name = root.tag
        if root_name == cls.cplex_solution_tag:
            mst1 = MSTReader.read_one_solution(root, mdl, mst_path)
            return None if mst1 is None else [mst1]
        elif root_name == cls.cplex_solutions_tag:
            return MSTReader.read_many_solutions(root, mdl, mst_path)
        else:
            mdl.fatal("Unexpected root element tag, expecting {0}|{1}, found: <{2}>"
                      , cls.cplex_solution_tag, cls.cplex_solutions_tag,
                      root_name)


def read_mst_file(mst_path, mdl):
    try:
        tree = xml_elt.parse(mst_path)
        root = tree.getroot()
        return MSTReader.read_root(root, mdl, mst_path)
    except xml_elt.ParseError as pex:
        mdl.error("XML error: {0!s} in file {1} - read aborted", pex, mst_path)
        # None is for errors
        return None




# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2016
# --------------------------------------------------------------------------

from six import iteritems

from docplex.mp.model import Model
from docplex.mp.aggregator import ModelAggregator
from docplex.mp.quad import VarPair
from docplex.mp.utils import is_number, is_iterable, generate_constant, \
    is_pandas_dataframe, is_pandas_series, is_numpy_matrix, is_scipy_sparse, is_ordered_sequence
from docplex.mp.constants import ComparisonType

from docplex.mp.compat23 import izip
from docplex.mp.error_handler import docplex_fatal
from docplex.mp.xcounter import update_dict_from_item_value

class AdvAggregator(ModelAggregator):
    def __init__(self, linear_factory, quad_factory):
        ModelAggregator.__init__(self, linear_factory, quad_factory)

    def _scal_prod_vars_all_different(self, terms, coefs):
        checker = self._checker
        if not is_iterable(coefs, accept_string=False):
            checker.typecheck_num(coefs)
            return coefs * self._sum_vars_all_different(terms)
        else:
            # coefs is iterable
            lcc_type = self.counter_type
            lcc = lcc_type()
            lcc_setitem = lcc_type.__setitem__
            number_validation_fn = checker.get_number_validation_fn()
            if number_validation_fn:
                for dvar, coef in izip(terms, coefs):
                    safe_coef = number_validation_fn(coef)
                    if safe_coef:
                        lcc_setitem(lcc, dvar, safe_coef)
            else:
                for dvar, coef in izip(terms, coefs):
                    if coef:  # zero test is much cheaper than a setitem
                        lcc_setitem(lcc, dvar, coef)

            return self._to_expr(qcc=None, lcc=lcc)

    def scal_prod_triple(self, left_terms, right_terms, coefs):
        used_coefs = None
        checker = self._model._checker

        if is_iterable(coefs, accept_string=False):
            used_coefs = coefs
        elif is_number(coefs):
            if coefs:
                used_coefs = generate_constant(coefs, count_max=None)
            else:
                return self.new_zero_expr()
        else:
            self._model.fatal("scal_prod_triple expects iterable or number as coefficients, got: {0!r}", coefs)

        if is_iterable(left_terms):
            used_left = checker.typecheck_var_seq(left_terms)
        else:
            checker.typecheck_var(left_terms)
            used_left = generate_constant(left_terms, count_max=None)

        if is_iterable(right_terms):
            used_right = checker.typecheck_var_seq(right_terms)
        else:
            checker.typecheck_var(right_terms)
            used_right = generate_constant(right_terms, count_max=None)

        if used_coefs is not coefs and used_left is not left_terms and used_right is not right_terms:
            # LOOK
            return left_terms * right_terms * coefs

        return self._scal_prod_triple(coefs=used_coefs, left_terms=used_left, right_terms=used_right)

    def _scal_prod_triple(self, coefs, left_terms, right_terms):
        # INTERNAL
        accumulated_ct = 0
        qcc = self.counter_type()
        lcc = self.counter_type()
        number_validation_fn = self._checker.get_number_validation_fn()
        for coef, lterm, rterm in izip(coefs, left_terms, right_terms):
            if coef:
                safe_coef = number_validation_fn(coef) if number_validation_fn else coef
                lcst = lterm.get_constant()
                rcst = rterm.get_constant()
                accumulated_ct += safe_coef * lcst * rcst
                for lv, lk in lterm.iter_terms():
                    for rv, rk in rterm.iter_terms():
                        coef3 = safe_coef * lk * rk
                        update_dict_from_item_value(qcc, VarPair(lv, rv), coef3)
                if rcst:
                    for lv, lk in lterm.iter_terms():
                        update_dict_from_item_value(lcc, lv, safe_coef * lk * rcst)
                if lcst:
                    for rv, rk in rterm.iter_terms():
                        update_dict_from_item_value(lcc, rv, safe_coef * rk * lcst)

        return self._to_expr(qcc, lcc, constant=accumulated_ct)

    def _scal_prod_triple_vars(self, coefs, left_terms, right_terms):
        # INTERNAL
        # assuming all arguments are iterable.
        dcc = self.counter_type
        qcc = dcc()
        number_validation_fn = self._checker.get_number_validation_fn()
        if number_validation_fn:
            for coef, lterm, rterm in izip(coefs, left_terms, right_terms):
                safe_coef = number_validation_fn(coef) if number_validation_fn else coef
                update_dict_from_item_value(qcc, VarPair(lterm, rterm), safe_coef)
        else:
            for coef, lterm, rterm in izip(coefs, left_terms, right_terms):
                update_dict_from_item_value(qcc, VarPair(lterm, rterm), coef)
        return self._to_expr(qcc=qcc)

    def _sumsq_vars_all_different(self, dvars):
        dcc = self._quad_factory.term_dict_type
        qcc = dcc()
        qcc_setitem = dcc.__setitem__
        for t in dvars:
            qcc_setitem(qcc, VarPair(t), 1)
        return self._to_expr(qcc=qcc)

    def _sumsq_vars(self, dvars):
        qcc = self._quad_factory.term_dict_type()
        for v in dvars:
            update_dict_from_item_value(qcc, VarPair(v), 1)
        return self._to_expr(qcc=qcc)

    def quad_matrix_sum(self, matrix, lvars, symmetric=False):
        # assume matrix is a NxN matrix
        # vars is a N-vector of variables
        dcc = self._quad_factory.term_dict_type
        qterms = dcc()

        gen_rows = self.generate_rows(matrix)

        for i, mrow in enumerate(gen_rows):
            vi = lvars[i]
            for j, k in enumerate(mrow):
                if k:
                    vj = lvars[j]
                    if i == j:
                        qterms[VarPair(vi)] = k
                    elif symmetric:
                        if i < j:
                            update_dict_from_item_value(qterms, VarPair(vi, vj), 2 * k)
                        elif i > j:
                            continue
                    else:
                        update_dict_from_item_value(qterms, VarPair(vi, vj), k)

        return self._to_expr(qcc=qterms)

    # noinspection PyUnusedLocal
    def _sparse_quad_matrix_sum(self, sp_coef_mat, lvars, symmetric=False):
        # assume matrix is a NxN matrix
        # vars is a N-vector of variables
        dcc = self._quad_factory.term_dict_type
        qterms = dcc()

        for e in range(sp_coef_mat.nnz):
            k = sp_coef_mat.data[e]
            if k:
                row = sp_coef_mat.row[e]
                col = sp_coef_mat.col[e]
                vi = lvars[row]
                vj = lvars[col]
                update_dict_from_item_value(qterms, VarPair(vi, vj), k)

        return self._to_expr(qcc=qterms)

    def vector_compare(self, left_exprs, right_exprs, sense):
        lfactory = self._linear_factory
        assert len(left_exprs) == len(right_exprs)
        cts = [lfactory._new_binary_constraint(left, sense, right) for left, right in izip(left_exprs, right_exprs)]
        return cts


# noinspection PyProtectedMember
class AdvModel(Model):
    """
    This class is a specialized version of the :class:`docplex.mp.model.Model` class with useful non-standard modeling
    functions.
    """
    _fast_settings = {'keep_ordering': False, 'checker': 'off', 'keep_all_exprs': False}

    def __init__(self, name=None, context=None, **kwargs):
        for k, v in iteritems(self._fast_settings):
            if k not in kwargs:
                # force fast settings if not present
                kwargs[k] = v

        Model.__init__(self, name=name, context=context, **kwargs)
        self._aggregator = AdvAggregator(self._lfactory, self._qfactory)

    def _prepare_constraint(self, ct, ctname, check_for_trivial_ct, arg_checker=None):
        # INTERNAL
        if ct is True:
            return False

        elif ct is False:
            # happens with sum([]) and constant e.g. sum([]) == 2
            msg = "Adding a trivially infeasible constraint"
            if ctname:
                msg += ' with name: {0}'.format(ctname)
            # analogous to 0 == 1, model is sure to fail
            self.fatal(msg)

        else:
            pass

        # --- name management ---
        if ctname:
            ct_name_map = self._cts_by_name
            if ct_name_map is not None:
                ct_name_map[ctname] = ct
            ct.name = ctname
        # ---
        return True

    def sumsq_vars(self, terms):
        return self._aggregator._sumsq_vars(terms)

    def sumsq_vars_all_different(self, terms):
        """
        Creates a quadratic expression by summing squares over a sequence.

        The variable sequence is a list or an iterator of variables.

        This method is faster than the standard summation of squares method due to the fact that it takes only variables and does not take expressions as arguments.

        :param terms: A list or an iterator on variables only, with no duplicates.

        :return: A quadratic expression or 0.

        Note:
           If the list or iterator is empty, this method returns zero.

        Note:
            To improve performance, the check for duplicates can be turned off by setting
            `checker='none'` in the `kwargs` of the :class:`docplex.mp.model.Model` object. As this argument
            turns off checking everywhere, it should be used with extreme caution.
        """
        var_seq = self._checker.typecheck_var_seq_all_different(terms)
        return self._aggregator._sumsq_vars_all_different(var_seq)

    def scal_prod_vars_all_different(self, terms, coefs):
        """
        Creates a linear expression equal to the scalar product of a list of decision variables and a sequence of coefficients.

        The variable sequence is a list or an iterator of variables.
        The coefficients can be either a list of numbers, an iterator over numbers, or a number.

        This method is faster than the standard generic scalar product method due to the fact that it takes only variables and does not take expressions as arguments.

        :param terms: A list or an iterator on variables only, with no duplicates.
        :param coefs: A list or an iterator on numbers, or a number.

        :return: A linear expression or 0.

        Note:
           If either list or iterator is empty, this method returns zero.

        Note:
            To improve performance, the check for duplicates can be turned off by setting
            `checker='none'` in the `kwargs` of the :class:`docplex.mp.model.Model` object. As this argument
            turns off checking everywhere, it should be used with extreme caution.
        """
        self._checker.check_ordered_sequence(arg=terms,
                                             caller='Model.scal_prod() requires a list of expressions/variables')
        var_seq = self._checker.typecheck_var_seq_all_different(terms)
        return self._aggregator._scal_prod_vars_all_different(var_seq, coefs)


    def quad_matrix_sum(self, matrix, dvars, symmetric=False):
        """
        Creates a quadratic expression equal to the quadratic form of a list of decision variables and
        a matrix of coefficients.

        This method sums all quadratic terms built by multiplying the [i,j]th coefficient in the matrix
        by the product of the i_th and j_th variables in `dvars`; in mathematical terms, the expression formed
        by x'Qx.

        :param matrix: A accepts either a list of lists of numbers, a numpy array, a pandas dataframe, or
            a scipy sparse matrix in COO format. T
            The resulting matrix must be square with size (N,N) where N is the number of variables.
        :param dvars: A list or an iterator on variables.
        :param symmetric: A boolean indicating whether the matrix is symmetric or not (default is False).
            No check is done.

        :return: An instance of :class:`docplex.mp.quad.QuadExpr` or 0.

        Note:
           The matrix must be square but not necessarily symmetric. The number of rows of the matrix must be equal
           to the size of the variable sequence.

           The symmetric flag only explores half of the matrix and doubles non-diagonal factors. No actual check is done.
           This flag has no effect on scipy sparse matrix.

        Example:
            `Model.quad_matrix_sum([[[1, 2], [3, 4]], [x, y])` returns the expression `x^2+4y^2+5x*yt`.
        """
        lvars = AdvModel._to_list(dvars, caller='Model.quad_matrix_sum')
        self._checker.typecheck_var_seq(lvars)
        if is_scipy_sparse(matrix):
            return self._aggregator._sparse_quad_matrix_sum(matrix, lvars, symmetric=symmetric)
        else:
            return self._aggregator.quad_matrix_sum(matrix, lvars, symmetric=symmetric)

    def scal_prod_triple(self, left_terms, right_terms, coefs):
        """
        Creates a quadratic expression from two lists of linear expressions and a sequence of coefficients.

        This method sums all quadratic terms built by multiplying the i_th coefficient by the product of the i_th
        expression in `left_terms` and the i_th expression in `right_terms`

        This method accepts different types of input for its arguments. The expression sequences can be either lists
        or iterators of objects that can be converted to linear expressions, that is, variables or linear expressions
        (but no quadratic expressions).
        The most usual case is variables.
        The coefficients can be either a list of numbers, an iterator over numbers, or a number.

        Example:
            `Model.scal_prod_triple([x, y], [z, t], [2, 3])` returns the expression `2xz + 3yt`.

        :param left_terms: A list or an iterator on variables or expressions.
        :param right_terms: A list or an iterator on variables or expressions.
        :param coefs: A list or an iterator on numbers or a number.

        :returns: An instance of :class:`docplex.mp.quad.QuadExpr` or 0.

        Note:
           If either list or iterator is empty, this method returns zero.
        """
        return self._aggregator.scal_prod_triple(left_terms=left_terms, right_terms=right_terms, coefs=coefs)

    def scal_prod_triple_vars(self, left_terms, right_terms, coefs):
        """
        Creates a quadratic expression from two lists of variables and a sequence of coefficients.

        This method sums all quadratic terms built by multiplying the i_th coefficient by the product of the i_th
        expression in `left_terms` and the i_th expression in `right_terms`

        This method is faster than the standard generic scalar quadratic product method due to the fact that it takes only variables and does not take expressions as arguments.

        Example:
            `Model.scal_prod_vars_triple([x, y], [z, t], [2, 3])` returns the expression `2xz + 3yt`.

        :param left_terms: A list or an iterator on variables.
        :param right_terms: A list or an iterator on variables.
        :param coefs: A list or an iterator on numbers or a number.

        :returns: An instance of :class:`docplex.mp.quad.QuadExpr` or 0.

        Note:
           If either list or iterator is empty, this method returns zero.
        """
        used_coefs = None
        checker = self._checker
        nb_non_iterables = 0

        if is_iterable(coefs, accept_string=False):
            used_coefs = coefs
        elif is_number(coefs):
            if coefs:
                used_coefs = generate_constant(coefs, count_max=None)
                nb_non_iterables += 1
            else:
                return self._aggregator.new_zero_expr()
        else:
            self.fatal("scal_prod_triple expects iterable or number as coefficients, got: {0!r}", coefs)

        if is_iterable(left_terms):
            used_left = checker.typecheck_var_seq(left_terms)
        else:
            nb_non_iterables += 1
            checker.typecheck_var(left_terms)
            used_left = generate_constant(left_terms, count_max=None)

        if is_iterable(right_terms):
            used_right = checker.typecheck_var_seq(right_terms)
        else:
            nb_non_iterables += 1
            checker.typecheck_var(right_terms)
            used_right = generate_constant(right_terms, count_max=None)

        if nb_non_iterables >= 3:
            return left_terms * right_terms * coefs
        else:
            return self._aggregator._scal_prod_triple_vars(left_terms=used_left,
                                                           right_terms=used_right, coefs=used_coefs)

    @classmethod
    def _to_list(cls, s, caller):
        if is_pandas_series(s):
            return s.tolist()
        elif is_ordered_sequence(s):
            return s
        else:
            docplex_fatal('{0} requires ordered sequences: lists, numpy array or Series, got: {1}', caller, type(s))
            return list(s)

    def matrix_constraints(self, coef_mat, dvars, rhs, sense='le'):
        """
        Creates a list of linear constraints
        from a matrix of coefficients, a sequence of variables, and a sequence of numbers.

        This method returns the list of constraints built from

            A.X <op> B

        where A is the coefficient matrix (of size (M,N)), X is the variable sequence (size N),
        and B is the sequence of right-hand side values (of size M).

        <op> is the comparison operator that defines the sense of the constraint. By default, this generates
        a 'less-than-or-equal' constraint.

        Example:
            `Model.scal_prod_vars_triple([x, y], [z, t], [2, 3])` returns the expression `2xz + 3yt`.

        :param coef_mat: A matrix of coefficients with M rows and N columns. This argument accepts
            either a list of lists of numbers, a `numpy` array with size (M,N), or a `scipy` sparse matrix.
        :param dvars: An ordered sequence of decision variables: accepts a Python list, `numpy` array,
            or a `pandas` series. The size of the sequence must match the number of columns in the matrix.
        :param rhs: A sequence of numbers: accepts a Python list, a `numpy` array,
            or a `pandas` series. The size of the sequence must match the number of rows in the matrix.
        :param sense: A constraint sense, accepts either a
            value of type `ComparisonType` or a string (e.g 'le', 'eq', 'ge').

        :returns: A list of linear constraints.

        Example:

            If A is a matrix of coefficients with 2 rows and 3 columns::

                    A = [[1, 2, 3],
                         [4, 5, 6]],
                    X = [x, y, z] where x, y, and z are decision variables (size 3), and

                    B = [100, 200], a sequence of numbers (size 2),

            then::

                `mdl.matrix_constraint(A, X, B, 'GE')` returns a list of two constraints
                [(x + 2y+3z <= 100), (4x + 5y +6z <= 200)].

        Note:
            If the dimensions of the matrix and variables or of the matrix and number sequence do not match,
            an error is raised.

        """
        checker = self._checker
        if is_pandas_dataframe(coef_mat) or is_numpy_matrix(coef_mat) or is_scipy_sparse(coef_mat):
            nb_rows, nb_cols = coef_mat.shape
        else:
            # a sequence of sequences
            a_mat = list(coef_mat)
            nb_rows = len(a_mat)
            nb_cols = None
            try:
                shared_len = None
                for r in a_mat:
                    checker.check_ordered_sequence(r, 'matrix_constraints')
                    r_len = len(r)
                    if shared_len is None:
                        shared_len = r_len
                    elif r_len != shared_len:
                        self.fatal('All columns should have same length found  {0} != {1}'.format(shared_len, r_len))
                nb_cols = shared_len if shared_len is not None else 0
            except AttributeError:
                self.fatal('All columns should have a len()')

        s_dvars = self._to_list(dvars, caller='Model.matrix-constraints()')
        s_rhs = self._to_list(rhs, caller='Model.matrix-constraints()')
        # check

        checker.typecheck_var_seq(s_dvars)
        for k in s_rhs:
            checker.typecheck_num(k)

        op = ComparisonType.parse(sense)
        # ---
        # check dimensions and whether to transpose or not.
        # ---
        nb_rhs = len(s_rhs)
        nb_vars = len(s_dvars)
        if (nb_rows, nb_cols) != (nb_rhs, nb_vars):
            self.fatal(
                'Dimension error, matrix is ({0},{1}), expecting ({3}, {2})'.format(nb_rows, nb_cols, nb_vars, nb_rhs))

        if is_scipy_sparse(coef_mat):
            return self._aggregator._sparse_matrix_constraints(coef_mat, s_dvars, s_rhs, op)
        else:
            return self._aggregator._matrix_constraints(coef_mat, s_dvars, s_rhs, op)

    def matrix_ranges(self, coef_mat, dvars, lbs, ubs):
        """
        Creates a list of range constraints
        from a matrix of coefficients, a sequence of variables, and two sequence of numbers.

        This method returns the list of range constraints built from

            L <= Ax <= U

        where A is the coefficient matrix (of size (M,N)), X is the variable sequence (size N),
        L and B are sequence of numbers (resp. the lower and upper bounds of the ranges) both with size M.


        :param coef_mat: A matrix of coefficients with M rows and N columns. This argument accepts
            either a list of lists of numbers, a `numpy` array with size (M,N), or a `scipy` sparse matrix.
        :param dvars: An ordered sequence of decision variables: accepts a Python list, `numpy` array,
            or a `pandas` series. The size of the sequence must match the number of columns in the matrix.
        :param lbs: A sequence of numbers: accepts a Python list, a `numpy` array,
            or a `pandas` series. The size of the sequence must match the number of rows in the matrix.
        :param ubs: A sequence of numbers: accepts a Python list, a `numpy` array,
            or a `pandas` series. The size of the sequence must match the number of rows in the matrix.

        :returns: A list of range constraints.

        Example::

            If A is a matrix of coefficients with 2 rows and 3 columns:

                    A = [[1, 2, 3],
                         [4, 5, 6]],
                    X = [x, y, z] where x, y, and z are decision variables (size 3), and

                    L = [101. 102], a sequence of numbers (size 2),
                    U = [201, 202]

            then::

                `mdl.range_constraints(A, X, L, U)` returns a list of two ranges
                [(101 <= x + 2y+3z <= 102), (201 <= 4x + 5y +6z <= 202)].

        Note:
            If the dimensions of the matrix and variables or of the matrix and number sequence do not match,
            an error is raised.

        """
        checker = self._checker
        if is_pandas_dataframe(coef_mat) or is_numpy_matrix(coef_mat) or is_scipy_sparse(coef_mat):
            nb_rows, nb_cols = coef_mat.shape
        else:
            # a sequence of sequences
            a_mat = list(coef_mat)
            nb_rows = len(a_mat)
            nb_cols = None
            try:
                shared_len = None
                for r in a_mat:
                    checker.check_ordered_sequence(r, 'matrix_constraints')
                    r_len = len(r)
                    if shared_len is None:
                        shared_len = r_len
                    elif r_len != shared_len:
                        self.fatal('All columns should have same length found  {0} != {1}'.format(shared_len, r_len))
                nb_cols = shared_len if shared_len is not None else 0
            except AttributeError:
                self.fatal('All columns should have a len()')

        s_dvars = self._to_list(dvars, caller='Model.range_constraints()')
        s_lbs = self._to_list(lbs, caller='Model.range_constraints()')
        s_ubs = self._to_list(ubs, caller='Model.range_constraints()')
        # check

        checker.typecheck_var_seq(s_dvars)
        checker.typecheck_num_seq(s_lbs, caller="AdvModel.matrix_ranges.lbs")
        checker.typecheck_num_seq(s_ubs, caller="AdvModel.matrix_ranges.ubs")

        # ---
        # check dimensions and whether to transpose or not.
        # ---
        nb_vars = len(s_dvars)
        nb_lbs = len(s_lbs)
        nb_ubs = len(s_ubs)
        if nb_lbs != nb_rows:
            self.fatal('Incorrect size for range lower bounds, expecting: {1}, got: {0},'.format(nb_lbs, nb_rows))
        if nb_ubs != nb_rows:
            self.fatal('Incorrect size for range upper bounds, expecting: {1}, got: {0}'.format(nb_ubs, nb_rows))
        if nb_cols != nb_vars:
            self.fatal(
                'Incorrect number of variables, expecting: {1}, got: {0},  matrix is ({0},{1})'.format(nb_vars, nb_cols,
                                                                                                       nb_rows,
                                                                                                       nb_cols))

        if is_scipy_sparse(coef_mat):
            return self._aggregator._sparse_matrix_ranges(coef_mat, s_dvars, s_lbs, s_ubs)
        else:
            return self._aggregator._matrix_ranges(coef_mat, s_dvars, s_lbs, s_ubs)

    def vector_compare(self, lhss, rhss, sense):
        l_lhs = self._to_list(lhss, caller='Model.vector.compare')
        l_rhs = self._to_list(rhss, caller='Model.vector.compare')
        if len(l_lhs) != len(l_rhs):
            self.fatal('Model.vector_compare() got sequences with different length, left: {0}, right: {1}'.
                       format(len(l_lhs), len(l_rhs)))
        ctsense = ComparisonType.parse(sense)
        return self._aggregator.vector_compare(l_lhs, l_rhs, ctsense)

    def vector_compare_le(self, lhss, rhss):
        return self.vector_compare(lhss, rhss, 'le')

    def vector_compare_ge(self, lhss, rhss):
        return self.vector_compare(lhss, rhss, 'ge')

    def vector_compare_eq(self, lhss, rhss):
        return self.vector_compare(lhss, rhss, 'eq')



import os
import re
from distutils.core import setup
import sys

required = ['requests',
            'six']
if ((sys.version_info[0]) < 3) or \
   ((sys.version_info[0] == 3) and (sys.version_info[1] < 2)):
    required.append('futures')

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    try:
        with open(os.path.join(HERE, *parts)) as f:
            return f.read()
    except:
        return None

readme = read('README.rst')
if readme is None:
    readme = 'DOcplex 2.15'

changelog = str(read('CHANGELOG.rst'))
if changelog is None:
    changelog = ''

ss = str(readme) + str(changelog)

setup(
    name='docplex',
    packages=['docplex',
               'docplex.cp',
               'docplex.cp.solver',
               'docplex.cp.cpo',
               'docplex.cp.fzn',
               'docplex.cp.lp',
               'docplex.mp',
               'docplex.mp.callbacks',
               'docplex.mp.internal',
               'docplex.mp.params',
               'docplex.mp.sktrans',
               'docplex.mp.sparktrans',
               'docplex.mp.worker',
               'docplex.util',
               'docplex.util.dods',
               'docplex.util.ws'],
    version = '2.15.194',  # replaced at build time
    description = 'The IBM Decision Optimization CPLEX Modeling for Python',
    author = 'The IBM Decision Optimization on Cloud team',
    author_email = 'dofeedback@wwpdl.vnet.ibm.com',
    long_description='%s\n' % ss,
    long_description_content_type='text/x-rst',
    url = 'https://onboarding-oaas.docloud.ibmcloud.com/software/analytics/docloud/',
    keywords = ['docloud', 'optimization', 'cplex', 'cpo'],
    license = 'Apache 2.0',
    install_requires=required,
    classifiers = ["Development Status :: 5 - Production/Stable",
                   "Intended Audience :: Developers",
                   "Intended Audience :: Information Technology",
                   "Intended Audience :: Science/Research",
                   "Operating System :: Unix",
                   "Operating System :: MacOS",
                   "Operating System :: Microsoft",
                   "Operating System :: OS Independent",
                   "Topic :: Scientific/Engineering",
                   "Topic :: Scientific/Engineering :: Mathematics",
                   "Topic :: Software Development :: Libraries",
                   "Topic :: System",
                   "Topic :: Other/Nonlisted Topic",
                   "License :: OSI Approved :: Apache Software License",
                   "Programming Language :: Python",
                   "Programming Language :: Python :: 2.7",
                   "Programming Language :: Python :: 3.6",
                   "Programming Language :: Python :: 3.7"
                   ],
)

print("** The documentation can be found here: https://github.com/IBMDecisionOptimization/docplex-doc")
print("** The examples can be found here: https://github.com/IBMDecisionOptimization/docplex-examples")



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore

from docplex.mp.utils import iter_emptyset
from docplex.mp.constants import ComparisonType


class Operand(object):
    __slots__ = ()

    def get_constant(self):
        return 0

    def is_constant(self):
        return False

    # --- basic subscription api
    def notify_used(self, user):
        pass

    def notify_unsubscribed(self, subscriber):
        pass

    # noinspection PyMethodMayBeStatic
    def is_in_use(self):
        return False

    # noinspection PyMethodMayBeStatic
    def is_shared(self):
        return False

    def notify_modified(self, event):
        pass

    def keep(self):
        return self
    # ---

    def resolve(self):
        # used for lazy expansions
        pass

    def get_linear_part(self):
        return self

    def __le__(self, rhs):
        return self._model._qfactory.new_xconstraint(lhs=self, rhs=rhs, comparaison_type=ComparisonType.LE)

    def __eq__(self, rhs):
        return self._model._qfactory.new_xconstraint(lhs=self, rhs=rhs, comparaison_type=ComparisonType.EQ)

    def __ge__(self, rhs):
        return self._model._qfactory.new_xconstraint(lhs=self, rhs=rhs, comparaison_type=ComparisonType.GE)


    le = __le__
    eq = __eq__
    ge = __ge__


class LinearOperand(Operand):
    # no ctor as used in multiple inheritance

    def unchecked_get_coef(self, dvar):
        raise NotImplementedError('unchecked_get_coef missing for class: {0}'.format(self.__class__))  # pragma: no cover

    def iter_variables(self):
        """
        Iterates over all variables in the expression.

        Returns:
            iterator: An iterator over all variables present in the operand.
        """
        for v, k in self.iter_terms():
            yield v

    def iter_terms(self):
        # iterates over all linear terms, if any
        return iter_emptyset()

    iter_sorted_terms = iter_terms

    def number_of_terms(self):
        return sum(1 for _ in self.iter_terms())

    @property
    def size(self):
        return self.number_of_terms()

    # noinspection PyMethodMayBeStatic
    def iter_quads(self):
        return iter_emptyset()

    def is_constant(self):
        # redefine this for subclasses.
        return False  # pragma: no cover

    def as_variable(self):
        # return a variable if the expression is actually one variable, else None
        return None

    def is_zero(self):
        return False

    def get_constant(self):
        return 0

    # no strict comparisons
    def __lt__(self, e):
        self.model.unsupported_relational_operator_error(self, "<", e)

    def __gt__(self, e):
        self.model.unsupported_relational_operator_error(self, ">", e)

    def __contains__(self, dvar):
        """Overloads operator `in` for an expression and a variable.

        :param: dvar (:class:`docplex.mp.linear.Var`): A decision variable.

        Returns:
            Boolean: True if the variable is present in the expression, else False.
        """
        return self.contains_var(dvar)

    def __ne__(self, rhs):
        return self._model._lfactory.new_neq_constraint(lhs=self, rhs=rhs)

    def contains_var(self, dvar):
        raise NotImplementedError  # pragma: no cover

    def lock_discrete(self):
        pass

    def is_discrete_locked(self):
        return False



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------


from __future__ import print_function


from docplex.mp.constants import ComparisonType
from docplex.mp.constr import LinearConstraint, RangeConstraint, QuadraticConstraint, PwlConstraint
from docplex.mp.environment import env_is_64_bit
from docplex.mp.mprinter import TextModelPrinter, _ExportWrapper, _NumPrinter
from docplex.mp.utils import fix_whitespace

from docplex.mp.format import LP_format
from itertools import chain
from six import iteritems, PY3
from docplex.mp.compat23 import izip


# gendoc: ignore

def _non_compliant_lp_name_stop_here(name):
    pass


class LPModelPrinter(TextModelPrinter):
    #_lp_re = re.compile(r"[a-df-zA-DF-Z!#$%&()/,;?@_`'{}|\"][a-zA-Z0-9!#$%&()/.,;?@_`'{}|\"]*")

    _lp_symbol_map = {ComparisonType.EQ: " = ",  # BEWARE NOT ==
                      ComparisonType.LE: " <= ",
                      ComparisonType.GE: " >= "}

    __new_line_sep = '\n'
    __expr_indent = ' ' * 6

    float_precision_32 = 9
    float_precision_64 = 12  #
    _nb_noncompliant_ids = 0
    _noncompliant_justifier = None

    def __init__(self, hide_user_names=False, **kwargs):
        nb_digits = self.float_precision_64 if env_is_64_bit() else self.float_precision_32
        TextModelPrinter.__init__(self,
                                  indent=1,
                                  comment_start='\\',
                                  hide_user_names=hide_user_names,
                                  nb_digits_for_floats=nb_digits)

        self._noncompliant_varname = None
        # specific printer for lp: do not print +inf/-inf inside constraints!
        self._lp_num_printer = _NumPrinter(nb_digits_for_floats=nb_digits,
                                           num_infinity=1e+20, pinf="1e+20", ninf="-1e+20")
        self._print_full_obj = kwargs.get('full_obj', False)

    def get_format(self):
        return LP_format

    def mangle_names(self):
        return TextModelPrinter.mangle_names(self) or self._noncompliant_varname

    def _print_ct_name(self, ct, name_map):
        lp_ctname = name_map.get(ct._index)
        indented = self._indent_level

        if lp_ctname is not None:
            ct_label = self._indent_space + lp_ctname + ':'
            indented += len(ct_label)
        else:
            ct_label = ''
        ct_indent_space = self._get_indent_from_level(indented)
        return ct_indent_space, ct_label

    def _print_binary_ct(self, wrapper, num_printer, var_name_map, binary_ct, _symbol_map=_lp_symbol_map,
                         allow_empty=False, force_first_sign=False):
        # ensure consistent ordering: left terms then right terms
        iter_diff_coeffs = binary_ct.iter_net_linear_coefs()
        self._print_expr_iter(wrapper, num_printer, var_name_map, iter_diff_coeffs,
                              allow_empty=True,  # when expr is empty print nothing, otherwise CPLEX crashes...!!!
                              force_first_plus=force_first_sign)
        wrapper.write(_symbol_map.get(binary_ct.sense, " ?? "), separator=False)
        wrapper.write(num_printer.to_string(binary_ct.cplex_num_rhs()), separator=False)

    def _print_ranged_ct(self, wrapper, num_printer, var_name_map, ranged_ct):
        exp = ranged_ct.expr
        (varname, rhs, _) = self._rangeData[ranged_ct]
        self._print_lexpr(wrapper, num_printer, var_name_map, exp)
        wrapper.write('-', separator=False)
        wrapper.write(varname)
        wrapper.write('=')
        wrapper.write(self._num_to_string(rhs))

    def _print_logical_ct(self, wrapper, num_printer, var_name_map, logical_ct,
                          logical_symbol):
        wrapper.write(self._var_print_name(logical_ct.binary_var))
        # rtc-39773 : write " = 1 -> " as one atomic symbol
        wrapper.write("= %d %s" % (logical_ct.active_value, logical_symbol))
        # wrapper.write("%d" % logical_ct.active_value)
        # wrapper.write(logical_symbol)
        self._print_binary_ct(wrapper, num_printer, var_name_map, logical_ct.linear_constraint)

    def _print_quadratic_ct(self, wrapper, num_printer, var_name_map, qct):
        q = self._print_qexpr_iter(wrapper, num_printer, var_name_map, qct.iter_net_quads())
        # force a '+' ?
        has_quads = q > 0
        self._print_binary_ct(wrapper, num_printer, var_name_map, qct, allow_empty=has_quads,
                              force_first_sign=has_quads)

    def _print_pwl_ct(self, wrapper, num_printer, var_name_map, pwl):
        """
        Prints a PWL ct in LP
        :param wrapper:
        :param pwl
        :return:
        """
        num2string_fn = num_printer.to_string
        wrapper.write('%s = %s' % (var_name_map[pwl.y._index], var_name_map[pwl.expr._index]))
        pwl_func = pwl.pwl_func
        pwl_def = pwl_func.pwl_def_as_breaks
        wrapper.write('%s' % num2string_fn(pwl_def.preslope))
        for pair in pwl_def.breaksxy:
            wrapper.write('(%s, %s)' % (num2string_fn(pair[0]), num2string_fn(pair[1])))
        wrapper.write('%s' % num2string_fn(pwl_def.postslope))

    def _print_constraint_label(self, wrapper, ct, name_map):
        if self._mangle_names:
            wrapper.set_indent('')
        else:
            indent_str, ct_label = self._print_ct_name(ct, name_map=name_map)
            wrapper.set_indent(indent_str)
            if ct_label is not None:
                wrapper.write(ct_label)

    def _print_constraint(self, wrapper, num_printer, var_name_map, ct):
        if isinstance(ct, PwlConstraint):
            # Pwl constraints are printed in a separate section (names 'PWL')
            return

        wrapper.begin_line()
        if isinstance(ct, LinearConstraint):
            self._print_constraint_label(wrapper, ct, name_map=self._linct_name_map)
            self._print_binary_ct(wrapper, num_printer, var_name_map, ct)
        elif isinstance(ct, RangeConstraint):
            self._print_constraint_label(wrapper, ct, name_map=self._linct_name_map)
            self._print_ranged_ct(wrapper, num_printer, var_name_map, ct)
        elif ct.is_logical():
            is_eq = ct.is_equivalence()
            logical_symbol = '<->' if is_eq else '->'
            self._print_constraint_label(wrapper, ct, name_map=self._lc_name_map)
            self._print_logical_ct(wrapper, num_printer, var_name_map, ct,
                                   logical_symbol=logical_symbol
                                   )
        elif isinstance(ct, QuadraticConstraint):
            self._print_constraint_label(wrapper, ct, name_map=self._qc_name_map)
            self._print_quadratic_ct(wrapper, num_printer, var_name_map, ct)
        else:
            ct.error("ERROR: unexpected constraint not printed: {0!s}".format(ct))  # pragma: no cover

        wrapper.flush(print_newline=True, restart_from_empty_line=True)

    def _print_pwl_constraint(self, wrapper, num_printer, var_name_map, ct):
        wrapper.begin_line()
        self._print_constraint_label(wrapper, ct, name_map=self._pwl_name_map)
        self._print_pwl_ct(wrapper, num_printer, var_name_map, ct)
        wrapper.flush(print_newline=True, restart_from_empty_line=True)

    def _print_var_block(self, wrapper, iter_vars, header):
        wrapper.begin_line()
        printed_header = False
        self_indent = self._indent_space
        for v in iter_vars:
            lp_name = self._var_print_name(v)
            if not printed_header:
                wrapper.newline()
                wrapper.write(header)
                printed_header = True
                wrapper.set_indent(self_indent)
                # Configure indent for next lines
                wrapper.flush(print_newline=True)
            wrapper.write(lp_name)
        if printed_header:
            wrapper.flush(print_newline=True)

    def _print_var_bounds(self, out, num_printer, varname, lb, ub, varname_indent=5 * ' ',
                          le_symbol='<=',
                          free_symbol='Free'):
        if lb is None and ub is None:
            # try to indent with space of '0 <= ', that is 5 space
            out.write(" %s %s %s\n" % (varname_indent, varname, free_symbol))
        elif lb is None:
            out.write(" %s %s %s %s\n" % (varname_indent, varname, le_symbol, num_printer.to_string(ub)))
        elif ub is None:
            out.write(" %s %s %s\n" % (num_printer.to_string(lb), le_symbol, varname))
        elif lb == ub:
            out.write(" %s %s %s %s\n" % (varname_indent, varname, "=", num_printer.to_string(lb)))
        else:
            out.write(" %s %s %s %s %s\n" % (num_printer.to_string(lb),
                                            le_symbol, varname, le_symbol,
                                            num_printer.to_string(ub)))

    TRUNCATE = 200

    def _notify_new_non_compliant_name(self, non_lp_name):
        _non_compliant_lp_name_stop_here(non_lp_name)
        self._nb_noncompliant_ids += 1
        if not self._noncompliant_justifier:
            self._noncompliant_justifier = non_lp_name

    def fix_name(self, mobj, prefix, local_index, hide_names):
        return LP_format.lp_name(mobj.name, prefix, local_index, hide_names,
                                 noncompliant_hook=self._notify_new_non_compliant_name)

    # def fix_name(self, mobj, prefix, local_index, hide_names):
    #     raw_name = mobj.get_name()
    #
    #     # anonymous constraints must be named in a LP (we follow CPLEX here)
    #     if hide_names or not raw_name: # or mobj.is_generated():
    #         return self._make_prefix_name(mobj, prefix, local_index, offset=1)
    #
    #     # # swap blanks with underscores
    #     ws_name = self._translate_chars(raw_name)
    #     if not self.is_lp_compliant(ws_name, fix_whitespace=False):
    #         if raw_name[0] in 'eE':
    #             # fixing eE non-LP names
    #             fixed_name = '_' + raw_name
    #             if self.is_lp_compliant(fixed_name, fix_whitespace=False):
    #                 return fixed_name
    #         # -- stats
    #         self._notify_new_non_compliant_name(raw_name)
    #         # --
    #         return self._make_prefix_name(mobj, prefix, local_index, offset=1)
    #     else:
    #
    #         # truncate if necessary, again this does nothing if name is too short
    #         return ws_name[:self.TRUNCATE]

    def _print_model_name(self, out, model):
        model_name = None
        if model.name:
            # make sure model name is ascii
            encoded = model.name.encode('ascii', 'backslashreplace')
            if PY3:  # pragma: no cover
                # in python 3, encoded is a bytes at this point. Make it a string again
                encoded = encoded.decode('ascii')
            model_name = encoded.replace('\\\\', '_').replace('\\', '_')
        printed_name = model_name or 'CPLEX'
        out.write("\\Problem name: %s\n" % printed_name)

    @staticmethod
    def is_lp_compliant(name, do_fix_whitespace=True):
        fixed_name = fix_whitespace(name) if do_fix_whitespace else name
        return LP_format.is_lp_compliant(fixed_name)

    @staticmethod
    def _is_injective(name_map):
        nb_keys = len(name_map)
        nb_different_names = len(set(name_map.values()))
        return nb_different_names == nb_keys

    @staticmethod
    def _has_sos_or_pwl_constraints(model):
        return model.number_of_sos > 0 or model.number_of_pwl_constraints > 0

    def _iter_completed_linear_obj_terms(self, model, objlin):
        obj_variables = set(v for v, _ in objlin.iter_terms())  # not sorted
        predeclared_variables = self._get_forced_predeclared_variables(model)
        variables_to_display = predeclared_variables | obj_variables
        # sort by index
        sorted_display_vars = list(variables_to_display)
        sorted_display_vars.sort(key=lambda dv: dv._index)
        for v in sorted_display_vars:
            yield v, objlin.unchecked_get_coef(v)

    def _iter_full_linear_obj_terms(self, model, objlin):
        # INTERNAL: print all variables and their coef in the linear part of the objective
        for v in model.iter_variables():
            yield v, objlin.unchecked_get_coef(v)

    def _get_forced_predeclared_variables(self, model):
        # compute predeclared variables
        predeclared_variables = set()
        for sos in model.iter_sos():
            for sos_var in sos.iter_variables():
                predeclared_variables.add(sos_var)
        for pwlc in model.iter_pwl_constraints():
            for pwv in pwlc.iter_extended_variables():
                predeclared_variables.add(pwv)
        return predeclared_variables

    def post_print_hook(self, model):
        nb_non_compliants = self._nb_noncompliant_ids
        if nb_non_compliants:
            try:
                model.warning('Some identifiers are not valid LP identifiers: %d (e.g.: "%s")',
                               nb_non_compliants, self._noncompliant_justifier)
            except UnicodeEncodeError:  # pragma: no cover
                model.warning('Some identifiers are not valid LP identifiers: %d (e.g.: "%s")',
                              nb_non_compliants, self._noncompliant_justifier.encode('utf-8'))

    def print_single_objective(self, wrapper, model, num_printer, var_name_map):
        wrapper.write(' obj:')
        objexpr = model.objective_expr

        if objexpr.is_quad_expr():
            objlin = objexpr.linear_part
        else:
            objlin = objexpr

        if self._print_full_obj:
            iter_linear_terms = self._iter_full_linear_obj_terms(model, objlin)
        elif self._has_sos_or_pwl_constraints(model):
            iter_linear_terms = self._iter_completed_linear_obj_terms(model, objlin)
        else:
            # write the linear part first
            # prints an expr to a stream
            iter_linear_terms = objlin.iter_sorted_terms()

        printed= self._print_expr_iter(wrapper, num_printer, var_name_map, iter_linear_terms,
                              allow_empty=True, accept_zero=True)

        if objexpr.is_quad_expr() and objexpr.has_quadratic_term():
            self._print_qexpr_obj(wrapper, num_printer, var_name_map,
                                  quad_expr=objexpr,
                                  force_initial_plus=printed)
            printed = True

        obj_offset = objexpr.get_constant()
        if obj_offset:
            if printed and obj_offset > 0:
                wrapper.write(u'+')
            wrapper.write(self._num_to_string(obj_offset))
        # ---
        wrapper.flush(print_newline=True)


    #  @profile
    def print_model_to_stream(self, out, model):
        # reset noncompliant stats
        self._nb_noncompliant_ids = 0
        self._noncompliant_justifier = None

        if not self._is_injective(self._var_name_map):
            # use indices to differentiate names
            import sys
            sys.__stdout__.write("DOcplex: refine variable names\n")
            k = 0
            for dv, lp_varname in iteritems(self._var_name_map):
                refined_name = "%s#%d" % (lp_varname, k)
                self._var_name_map[dv] = refined_name
                k += 1

        TextModelPrinter.prepare(self, model)
        self_num_printer = self._lp_num_printer
        var_name_map = self._var_name_map
        line_size = model.lp_line_length
        wrapper = _ExportWrapper(out, indent_str=self.__expr_indent,line_width=line_size)

        self._print_signature(out)
        self._print_encoding(out)
        self._print_model_name(out, model)
        self._newline(out)

        # ---  print objective
        if model.has_multi_objective():

            out.write(model.objective_sense.name)
            out.write(' multi-objectives')
            self._newline(out)
            env = model.environment
            if env.has_cplex and env.cplex_version.startswith("12.9."):
                mobj_indent = 5
                def make_default_obj_name(o_):
                    return 'obj%d' % o_ if o_ > 0 else 'obj'
            else:
                mobj_indent = 7 # length of 'obj1: '
                def make_default_obj_name(o_):
                    return 'obj%d' % (o_+1)

            wrapper.set_indent(mobj_indent * ' ') # length of 'obj1: '
            for o, ot in enumerate(model.iter_multi_objective_tuples()):
                expr, prio, w, abstol, reltol, oname = ot
                name = oname or make_default_obj_name(o)
                obj_label = '%s%s:' % (' ', name)
                wrapper.write(obj_label)
                wrapper.write('Priority=%g Weight=%g AbsTol=%g RelTol=%g' % (prio, w, abstol, reltol))
                wrapper.flush()
                # print the expression and its constant
                printed = self._print_expr_iter(wrapper, self_num_printer, var_name_map, expr.iter_terms(),
                                   allow_empty=True, accept_zero=True)
                obj_offset = expr.get_constant()
                if obj_offset:
                    if printed and obj_offset > 0:
                        wrapper.write(u'+')
                    wrapper.write(self_num_printer.to_string(obj_offset))

                wrapper.flush(restart_from_empty_line=True)

        else:
            out.write(model.objective_sense.name)
            self._newline(out)
            self.print_single_objective(wrapper, model, self_num_printer, var_name_map)

        # wrapper.write(' obj:')
        # objexpr = model.objective_expr
        #
        # if objexpr.is_quad_expr():
        #     objlin = objexpr.linear_part
        # else:
        #     objlin = objexpr
        #
        # if self._print_full_obj:
        #     iter_linear_terms = self._iter_full_linear_obj_terms(model, objlin)
        # elif self._has_sos_or_pwl_constraints(model):
        #     iter_linear_terms = self._iter_completed_linear_obj_terms(model, objlin)
        # else:
        #     # write the linear part first
        #     # prints an expr to a stream
        #     iter_linear_terms = objlin.iter_sorted_terms()
        #
        # printed= self._print_expr_iter(wrapper, self_num_printer, var_name_map, iter_linear_terms,
        #                       allow_empty=True, accept_zero=True)
        #
        # if objexpr.is_quad_expr() and objexpr.has_quadratic_term():
        #     self._print_qexpr_obj(wrapper, self_num_printer, var_name_map,
        #                           quad_expr=objexpr,
        #                           force_initial_plus=printed)
        #     printed = True
        #
        # obj_offset = objexpr.get_constant()
        # if obj_offset:
        #     if printed and obj_offset > 0:
        #         wrapper.write(u'+')
        #     wrapper.write(self._num_to_string(obj_offset))
        # # ---
        #
        # wrapper.flush(print_newline=True)

        out.write("Subject To\n")

        for ct in model.iter_constraints():
            self._print_constraint(wrapper, self_num_printer, var_name_map, ct)
        for lct in model.iter_implicit_equivalence_cts():
            wrapper.begin_line(True)
            self._print_logical_ct(wrapper, self_num_printer, var_name_map, lct, '<->')
            wrapper.flush(restart_from_empty_line=True)

        # lazy constraints
        self.print_linear_constraint_section(out, wrapper, model.iter_lazy_constraints(),
                                             "Lazy Constraints", ct_prefix="l", var_name_map=var_name_map,
                                             ctname_map=self._lzc_name_map)
        self.print_linear_constraint_section(out, wrapper, model.iter_user_cut_constraints(),
                                             "User Cuts", ct_prefix="u", var_name_map=var_name_map,
                                             ctname_map=self._ucc_name_map)

        out.write("\nBounds\n")
        symbolic_num_printer = self._num_printer
        print_var_bounds_fn = self._print_var_bounds
        var_print_name_fn = self._var_print_name
        for dvar in model.iter_variables():
            lp_varname = var_print_name_fn(dvar)
            var_lb = dvar._get_lb()
            var_ub = dvar._get_ub()
            if dvar.is_binary():
                print_var_bounds_fn(out, self_num_printer, lp_varname, var_lb, var_ub)
            else:

                free_lb = dvar.has_free_lb()
                free_ub = dvar.has_free_ub()
                if free_lb and free_ub:
                    print_var_bounds_fn(out, self_num_printer, lp_varname, lb=None, ub=None)
                elif free_ub:
                    # print only nonzero lbs
                    if var_lb:
                        print_var_bounds_fn(out, symbolic_num_printer, lp_varname, var_lb, ub=None)

                else:
                    # save the lb if is zero
                    printed_lb = None if 0 == var_lb else var_lb
                    print_var_bounds_fn(out, symbolic_num_printer, lp_varname, lb=printed_lb, ub=var_ub)

        # add ranged cts vars
        for rng in model.iter_range_constraints():
            (varname, _, rngval) = self._rangeData[rng]
            self._print_var_bounds(out, self_num_printer, varname, rngval, 0)

        iter_semis = chain(model.iter_semicontinuous_vars(), model.iter_semiinteger_vars())

        self._print_var_block(wrapper, model.iter_binary_vars(), 'Binaries')
        self._print_var_block(wrapper, chain(model.iter_integer_vars(), model.iter_semiinteger_vars()), 'Generals')
        self._print_var_block(wrapper, iter_semis, 'Semi-continuous')
        self._print_sos_block(wrapper, model)
        self._print_pwl_block(wrapper, model, self_num_printer, var_name_map)
        out.write("End\n")

    def print_linear_constraint_section(self, out, wrapper, ct_iter, section_label, ct_prefix, var_name_map, ctname_map):
        ct_count = 0
        self_indent = self._indent_space
        self_num_printer = self._lp_num_printer
        for lct in ct_iter:
            if not ct_count:
                out.write("\n{0}\n".format(section_label))
            ct_count += 1
            ct_name = ctname_map.get(lct) or '%s%d' % (ct_prefix, ct_count)
            ct_label = '%s%s:' % (self_indent, ct_name)
            wrapper.set_indent(self._get_indent_from_level(len(ct_label)+1))

            # print a normal linear constraint
            wrapper.write(ct_label)
            self._print_binary_ct(wrapper, self_num_printer, var_name_map, lct)
            wrapper.flush(print_newline=True, restart_from_empty_line=True)

    def _print_sos_block(self, wrapper, mdl):
        if mdl.number_of_sos > 0:
            wrapper.write('SOS')
            wrapper.flush(print_newline=True)
            name_fn = self._var_print_name
            for s, sos in enumerate(mdl.iter_sos(), start=1):
                # always print a name string, s1, s2, ... if no name
                sos_name = sos.get_name() or 's%d' % s
                wrapper.write('%s:' % sos_name)
                wrapper.write('S%d ::' % sos.sos_type.value)  # 1 or 2
                ranks = sos.weights
                # noinspection PyArgumentList
                for rank, sos_var in izip(ranks, sos._variables):
                    wrapper.write('%s : %d' % (name_fn(sos_var), rank))
                wrapper.flush(print_newline=True)

    def _print_pwl_block(self, wrapper, mdl, self_num_printer, var_name_map):
        if mdl.number_of_pwl_constraints > 0:
            wrapper.write('Pwl')
            wrapper.flush(print_newline=True)
            # name_fn = self._var_print_name
            for pwl in mdl.iter_pwl_constraints():
                self._print_pwl_constraint(wrapper, self_num_printer, var_name_map, pwl)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
This module implements a simulator of model solver.
It is mainly used for testing.
"""

from docplex.cp.expression import *
from docplex.cp.solution import *
import docplex.cp.solver.solver as solver
from docplex.cp.utils import *

import random

###############################################################################
##  Constants
###############################################################################

# Max value used by solution simulator
MAX_SIMULATED_VALUE = 10000

# Max number of solutions when using next() iterator
MAX_ITERATED_SOLUTIONS = 6


###############################################################################
##  Public classes
###############################################################################

class CpoSolverSimulatorFail(solver.CpoSolverAgent):
    """ CPO solver simulator that always fail (status unfeasible) """
    
    def __init__(self, solver, params, context):
        """ Create a solver simulator

        Args:
            solver:   Parent solver
            params:   Solving parameters
            context:  Solver context
        """
        super(CpoSolverSimulatorFail, self).__init__(solver, params, context)


    def solve(self):
        """ Solve the model.

        Returns:
            Model solution expressed as CpoModelSolution
        """
        # Warn about simulator
        print("WARNING: Solver simulator always returns fail")

        # Build fake infeasible solution
        msol = CpoSolveResult(self.model)
        msol.solve_status = SOLVE_STATUS_INFEASIBLE
        return msol


    def refine_conflict(self):
        """ This method identifies a minimal conflict for the infeasibility of the current model.

        Given an infeasible model, the conflict refiner can identify conflicting constraints and variable domains
        within the model to help you identify the causes of the infeasibility.
        In this context, a conflict is a subset of the constraints and/or variable domains of the model
        which are mutually contradictory.
        Since the conflict is minimal, removal of any one of these constraints will remove that
        particular cause for infeasibility.
        There may be other conflicts in the model; consequently, repair of a given conflict
        does not guarantee feasibility of the remaining model.

        Returns:
            Conflict result,
            object of class :class:`~docplex.cp.solution.CpoRefineConflictResult`.
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        # Warn about simulator
        print("WARNING: Solver simulator always returns fail")
        # Build fake infeasible solution
        msol = CpoRefineConflictResult(self.model)
        msol.solve_status = SOLVE_STATUS_INFEASIBLE
        return msol



class CpoSolverSimulatorRandom(solver.CpoSolverAgent):
    """ CPO solver simulator that generates a random solution """

    def __init__(self, solver, params, context):
        """ Create a solver simulator

        Args:
            solver:   Parent solver
            params:   Solving parameters
            context:  Solver context
        """
        super(CpoSolverSimulatorRandom, self).__init__(solver, params, context)
        self.max_iterator_solutions = random.randint(0, MAX_ITERATED_SOLUTIONS)
        self.solution_count = 0


    def solve(self):
        """ Solve the model.

        Returns:
            Model solve result (object of class CpoSolveResult)
        """
        # Warn about simulator
        print("WARNING: Solver simulator returns a random solution")

        # Force generation of CPO format if required (for testing purpose only)
        if self.context.create_cpo:
            self._get_cpo_model_string()

        # Build fake feasible solution
        return self._generate_solution()


    def refine_conflict(self):
        """ This method identifies a minimal conflict for the infeasibility of the current model.

        Given an infeasible model, the conflict refiner can identify conflicting constraints and variable domains
        within the model to help you identify the causes of the infeasibility.
        In this context, a conflict is a subset of the constraints and/or variable domains of the model
        which are mutually contradictory.
        Since the conflict is minimal, removal of any one of these constraints will remove that
        particular cause for infeasibility.
        There may be other conflicts in the model; consequently, repair of a given conflict
        does not guarantee feasibility of the remaining model.

        Returns:
            Conflict result,
            object of class :class:`~docplex.cp.solution.CpoRefineConflictResult`.
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        # Build fake conflict
        csol = CpoRefineConflictResult(self.model)
        vars = self.model.get_all_variables()
        if len(vars) > 0:
            for i in range(1 + (len(vars) // 2)):
                csol.member_variables.append(vars[random.randint(0, len(vars) - 1)])
        ctrs = self.model.get_all_expressions()
        if len(ctrs) > 0:
            for i in range(1 + (len(ctrs) // 2)):
                csol.member_constraints.append(ctrs[random.randint(0, len(ctrs) - 1)][0])

        return csol


    def start_search(self):
        """ Start a new search. Solutions are retrieved using method search_next().

        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        print("WARNING: Solver simulator returns a random list of solutions")
        # Force generation of CPO format if required (for testing purpose only)
        if self.context.create_cpo:
            self._get_cpo_model_string()
        self.solution_count = 0


    def search_next(self):
        """ Search the next available solution.

        Returns:
            Next solution (object of class CpoSolveResult)
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        if self.solution_count >= self.max_iterator_solutions:
            # Generate last solution
            msol = CpoSolveResult(self.model)
            msol.solve_status = SOLVE_STATUS_FEASIBLE
            msol.fail_status = FAIL_STATUS_SEARCH_COMPLETED
            return msol
        else:
           # Generate next solution
           self.solution_count += 1
           return self._generate_solution()


    def end_search(self):
        """ End current search.

        Returns:
            Last (fail) model solution with last solve information (type CpoSolveResult)
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        pass


    def _generate_solution(self):
        """ Generate a random solution complying to the model

        Returns:
            Random model solution expressed as CpoModelSolution
        """
        # Build fake feasible solution
        ssol = CpoSolveResult(self.model)
        ssol.solve_status = SOLVE_STATUS_FEASIBLE
        msol = ssol.solution

        # Generate objective
        x = self.model.get_optimization_expression()
        if x is not None:
            # Determine number of values
            x = x.children[0]
            if x.type.is_array:
                nbval = len(x.value)
            else:
                nbval = 1
            ovals = []
            for i in range(nbval):
                ovals.append(random.randint(0, MAX_SIMULATED_VALUE))
            msol.objective_values = ovals

        # Generate a solution for each variable
        allvars = self.model.get_all_variables()
        seqvars = []
        for var in allvars:
            if isinstance(var, CpoIntVar):
                vsol = CpoIntVarSolution(var, _random_value_in_complex_domain(var.get_domain()))
                msol.add_var_solution(vsol)

            elif isinstance(var, CpoIntervalVar):
                # Generate presence
                if var.is_absent():
                    present = False
                elif var.is_present():
                    present = True
                else:
                    present = (random.random() > 0.3)
                # Generate start and end
                dom = _common_interval_domain(var.get_start(), var.get_end())
                if dom[0] == dom[1]:
                    start = end = dom[0]  # Allow zero-length interval ?
                else:
                    start = end = 0
                    while start >= end:
                        start = _random_value_in_interval_domain(dom)
                        end = _random_value_in_interval_domain(dom)
                # Generate size
                size = _random_value_in_interval_domain(var.get_size())
                # Add variable to solution
                vsol = CpoIntervalVarSolution(var, present, start, end, size)
                msol.add_var_solution(vsol)

            elif isinstance(var, CpoStateFunction):
                # Build list of steps
                lsteps = []
                cstart = 0
                while cstart < MAX_SIMULATED_VALUE:
                    size = random.randint(1, MAX_SIMULATED_VALUE / 10)
                    lsteps.append((cstart, cstart + size, random.randint(0, 10)))
                    cstart += size
                vsol = CpoStateFunctionSolution(var, lsteps)
                msol.add_var_solution(vsol)

            elif isinstance(var, CpoSequenceVar):
                seqvars.append(var)

        # Generate a solution for sequence variables (done after all vars to access their solutions)
        for var in seqvars:
            # Build sequence or results
            lvres = []
            for v in var.get_interval_variables():
                lvres.append(msol.get_var_solution(v))
            random.shuffle(lvres)
            vsol = CpoSequenceVarSolution(var, lvres)
            msol.add_var_solution(vsol)

        # Return
        return ssol


###############################################################################
##  Private methods
###############################################################################

def _random_value_in_interval_domain(dom):
    """ Determine a random integer value in a domain expressed as a single interval

    Args:
        dom:  Interval domain (couple of values)
    Returns:
        Random value in this domain
    """
    return random.randint(dom[0], min(dom[1], dom[0] + MAX_SIMULATED_VALUE))

def _random_value_in_complex_domain(dom):
    """ Determine a random integer value in a domain

    Args:
        dom:  Value domain, list of integers or interval tuples
    Returns:
        Random value in this domain
    """
    # First select a domain element
    dlm = dom[random.randint(0, len(dom) - 1)]
    if is_int(dlm):
        return dlm
    return _random_value_in_interval_domain(dlm)

def _common_interval_domain(dom1, dom2):
    """ Determine the interval domain that is common to two interval domains

    Args:
        dom1:  First interval domain
        dom2:  Second interval domain
    Returns:
        Common domain
    """
    return max(dom1[0], dom2[0]), min(dom1[1], dom2[1])




# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

import json


class JSONSolutionHandler(object):
    '''A class that provides utilities to process the JSON solutions from DOcplexcloud
    '''

    # json keys
    JSON_LINEAR_CTS_KEY = 'linearConstraints'

    def __init__(self, solution_string, has_solution=None):
        '''Initialize a new JSONSolutionHandler

        This handler is initialized with a json fragment with the CPLEXSolution.

        Args:
            solution_string: The json text containing a CPLEXSolution
            has_solution: if set to True or False, forces the has_solution status. If None, this
                is True if json is not None.
        '''
        self.has_solution = bool(solution_string) if has_solution is None else has_solution
        if solution_string is not None:
            self.json = json.loads(solution_string, parse_constant='utf-8')['CPLEXSolution']
        else:
            self.json = None
        # used to store data from the json
        self.__vars = None

    def _getvars(self):
        if self.__vars is None:
            self.__vars = self.json.get('variables', [])
        return self.__vars

    def get_variable_attr_map(self, json_attr_name):
        assert json_attr_name
        if not self.json:
            return {}
        else:
            all_vars = self._getvars()
            attr_map = {int(v['index']): float(v[json_attr_name]) for v in all_vars}
            return attr_map

    def cplex_index_name_map(self):
        json_res = self.json
        if json_res:
            return {int(v['index']): v['name'] for v in self._getvars()}
        else:
            return {}

    def variable_results(self):
        if not self.json:
            return {}, {}
        else:
            all_vars = self._getvars()
            value_map = {int(v['index']): float(v['value']) for v in all_vars}
            if self.is_mip():
                rc_map = {}
            else:
                try:
                    rc_map = {int(v['index']): float(v['reducedCost']) for v in all_vars}
                except KeyError:
                    rc_map = {}
            return value_map, rc_map

    def constraint_results(self):
        if not self.json or self.is_mip():
            return {}, {}
        else:
            lincst_key = self.JSON_LINEAR_CTS_KEY
            if lincst_key in self.json:
                all_linear_cts = self.json[lincst_key]
                try:
                    dual_map = {int(v['index']): float(v['dual']) for v in all_linear_cts}
                except KeyError:
                    dual_map = {}
                try:
                    slack_map = {int(v['index']): float(v['slack']) for v in all_linear_cts}
                except KeyError:
                    slack_map = {}
            else:
                dual_map = slack_map = {}
            return dual_map, slack_map

    def constraint_slacks(self, linkey=JSON_LINEAR_CTS_KEY):
        self_json = self.json
        slack_map = {}
        if self_json and linkey in self_json:
            all_linear_cts = self_json[linkey]
            try:
                slack_map = {int(v['index']): float(v['slack']) for v in all_linear_cts}
            except KeyError:
                pass
        return slack_map

    def is_mip(self):
        return self.json and self.json['header']['solutionMethodString'] == 'mip'

    def _check_nonempty_json(self):
        if not self.json:
            raise ValueError("* empty JSON result!")

    def get_status_id(self):
        self._check_nonempty_json()
        return int(self.json['header']['solutionStatusValue'])

    def get_objective(self):
        self._check_nonempty_json()
        return float(self.json['header']['objectiveValue'])

    def variable_values(self):
        return self.get_variable_attr_map('value')

    def variable_reduced_costs(self):
        return self.get_variable_attr_map('reducedCost')



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

"""
IBM Decision Optimization CPLEX Modeling for Python - Constraint Programming

This package contains a Python API allowing to build Constraint Programming
models and their solving using the Decision Optimization cloud services.
"""

import platform
import sys
import docplex.version as dcpv
import warnings

# Check platform system
psyst = platform.system()
if psyst.lower() not in ('darwin', 'linux', 'windows', 'microsoft', 'aix'):
    msg = "DOcplex.CP is supported on Linux, Windows, Darwin and AIX, not on '{}'. Use it at your own risk.".format(psyst)
    warnings.warn(msg, RuntimeWarning)

# Check version of Python
pv = sys.version_info
if (pv < (2, 7)) or ((pv[0] == 3) and (pv < (3, 4) or pv >= (3, 8))):
    msg = "DOcplex.CP is supported by Python versions 2.7.9+, 3.4.x, to 3.7.x, not '{}'. Use it at your own risk."\
        .format('.'.join(str(x) for x in pv))
    warnings.warn(msg, RuntimeWarning)

# Set version information
__version_info__ = (dcpv.docplex_version_major, dcpv.docplex_version_minor, dcpv.docplex_version_micro)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore
"""
This is the docplex package.
"""



from docplex.version import docplex_version_string as __version__
from docplex.version import docplex_version_major, docplex_version_minor, docplex_version_micro
__version_info__ = (docplex_version_major, docplex_version_minor, docplex_version_micro)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# coding=utf-8
# ------------------------------
from __future__ import print_function
import sys

import six

from docplex.mp.numutils import _NumPrinter

# gendoc: ignore


class ModelPrinter(object):
    ''' Generic Printer code.
    '''

    def __init__(self):
        pass

    def get_format(self):
        """
        returns the Format object
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    def extension(self):
        """
        :return: the extension of the format
        """
        return self.get_format().extension

    def printModel(self, mdl, out=None):
        """ Generic method.
            If passed with a string, uses it as a file name
            if None is passed, uses standard output.
            else assume a stream is passed and try it
        """
        mdl.resolve()
        if out is None:
            # prints on standard output
            self.print_model_to_stream(sys.stdout, mdl)
        elif isinstance(out, six.string_types):
            # a string is interpreted as a path name
            ext = self.extension()
            path = out if out.endswith(ext) else out + ext
            # SAv format requires binary mode!
            write_mode = "wb" if self.get_format().is_binary else "w"
            with open(path, write_mode) as of:
                self.print_model_to_stream(of, mdl)
                # print("* file: %s overwritten" % path)
        else:
            try:
                self.print_model_to_stream(out, mdl)
            except AttributeError:  # pragma: no cover
                pass  # pragma: no cover
                # stringio will raise an attribute error here, due to with
                # print("Cannot use this an output: %s" % str(out))
        self.post_print_hook(mdl)

    def print_model_to_stream(self, out, mdl):
        raise NotImplementedError  # pragma: no cover

    def get_var_name_encoding(self):  # pragma: no cover
        return None  # default is no encoding

    def post_print_hook(self, model):
        # this method is called after printing the model
        # can be redefined for post-print reporting
        pass


class _DisambiguateError(Exception):
    pass


# noinspection PyAbstractClass
class TextModelPrinter(ModelPrinter):
    DEFAULT_ENCODING = "ENCODING=ISO-8859-1"

    def __init__(self, comment_start, indent=1,
                 hide_user_names=False,
                 nb_digits_for_floats=3,
                 encoding=DEFAULT_ENCODING,
                 sort_variable_names=False):
        ModelPrinter.__init__(self)
        # should be elsewhere
        self.true_infinity = float('inf')

        self.line_width = 79
        # noinspection PyArgumentEqualDefault

        self._comment_start = comment_start
        self._mangle_names = hide_user_names
        self._encoding = encoding  # None is a valid value, in which case no encoding is printed
        # -----------------------
        # TODO: refactor these maps as scope objects...
        self._var_name_map = {}
        self._linct_name_map = {}  # linear constraints
        self._lc_name_map = {}  # indicators have a seperate index space.
        self._qc_name_map = {}
        self._pwl_name_map = {}
        self._lzc_name_map = {}  # lazy constraints
        self._ucc_name_map = {}  # user cut constraints
        # ------------------------

        self._rangeData = {}
        self._num_printer = _NumPrinter(nb_digits_for_floats)
        self._indent_level = indent
        self._indent_space = ' ' * indent
        self._indent_map = {1: ' '}
        
        self.sort_variable_names = sort_variable_names

        # which translate_method to use
        if six.PY2:
            self._translate_chars = self._translate_chars2
        else:  # pragma: no cover
            self._translate_chars = self._translate_chars3

    def _get_indent_from_level(self, level):
        cached_indent = self._indent_map.get(level)
        if cached_indent is None:
            indent = ' ' * level
            self._indent_map[level] = indent
            return indent
        else:
            return cached_indent

    @property
    def nb_digits_for_floats(self):  # pragma: no cover
        return self._num_printer.precision


    def mangle_names(self):
        """
        Actually used to decide whether to encryupt or noyt
        :return:
        """
        return self._mangle_names

    def set_mangle_names(self, mangled):
        self._mangle_names = mangled

    def is_mangling_names(self):
        return self._mangle_names

    def _print_line_comment(self, out, comment_text):
        out.write("%s %s\n" % (self._comment_start, comment_text))

    def _print_encoding(self, out):
        """
        prints the file encoding
        :return:
        """
        if self._encoding:
            self._print_line_comment(out, self._encoding)

    def _print_model_name(self, out, mdl):
        """ Redefine this method to print the model name, if necessary
        :param mdl: the model to be printed
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    def _print_signature(self, out):
        """
        Prints a signature message denoting this file comes from Python Modeling Layer
        :return:
        """
        self._print_line_comment(out, "This file has been generated by DOcplex")

    def _newline(self, out, nb_lines=1):
        for _ in range(nb_lines):
            out.write("\n")

    disambiguate_try_max = 1000

    def _disambiguate(self, candidate_name, names, mobj, prefix, local_idx, try_max=disambiguate_try_max):
        # candidate_name is already in names
        # we coin successive names with index until the suffixed name is no longer in names
        k = 1
        disambiguate_fmt = '%s##%d'
        cur_name = candidate_name

        while cur_name in names:
            if k >= try_max:
                # giving up
                return self._make_prefix_name(mobj, prefix, local_idx)
            cur_name = disambiguate_fmt % (candidate_name, k)
            k+= 1
        # --
        return cur_name

    def _precompute_name_dict(self, mobj_seq, prefix, indexerfn_=None):

        if indexerfn_ is None:
            indexerfn = lambda mo: mo.index
        else:
            indexerfn = indexerfn_
        fixed_name_dir = {}
        all_names = set()
        hide_names = self.mangle_names()

        for local_index, mobj in enumerate(mobj_seq):
            fixed_name = self.fix_name(mobj, prefix, local_index, hide_names)
            if fixed_name:
                if fixed_name in all_names:
                    # to disambiguate, start with name#index, then add ##1,2,3,
                    seed_name = '%s#%d' % (fixed_name, mobj._index)
                    fixed_name = self._disambiguate(seed_name, all_names, mobj, prefix, local_index)
                fixed_name_dir[indexerfn(mobj)] = fixed_name
                all_names.add(fixed_name)

        return fixed_name_dir

    def _precompute_name_dict_by_obj(self, mobj_seq, prefix):
        return self._precompute_name_dict(mobj_seq, prefix, indexerfn_=lambda _x: _x)

    def _num_to_string(self, num):
        # INTERNAL
        return self._num_printer.to_string(num)

    def prepare(self, model):
        self._var_name_map = self._precompute_name_dict(model.iter_variables(), prefix='x')
        self._linct_name_map = self._precompute_name_dict(model.iter_linear_constraints(), prefix='c')
        self._lc_name_map = self._precompute_name_dict(model.iter_logical_constraints(), prefix='lc')
        self._qc_name_map = self._precompute_name_dict(model.iter_quadratic_constraints(), prefix='qc')
        self._pwl_name_map = self._precompute_name_dict(model.iter_pwl_constraints(), prefix='pwl')
        self._lzc_name_map = self._precompute_name_dict_by_obj(model.iter_lazy_constraints(), prefix='l')
        self._ucc_name_map = self._precompute_name_dict_by_obj(model.iter_user_cut_constraints(), prefix='u')

        self._rangeData = {}
        for rng in model.iter_range_constraints():
            # precompute data for ranges
            # 1 name ?
            # 2 rhs is lb - constant
            # 3 bounds are (0, ub-lb)
            varname = 'Rg%s' % self.linearct_print_name(rng)
            rhs = rng.cplex_num_rhs()
            rngval = rng.cplex_range_value(do_raise=False)
            self._rangeData[rng] = (varname, rhs, rngval)

    def _var_print_name(self, dvar):
        # INTERNAL
        return self._var_name_map[dvar._index]

    def get_name_to_var_map(self, model):
        # INTERNAL
        name_to_var_map = {}
        for v in model.iter_variables():
            name_to_var_map[self._var_name_map[v._index]] = v
        return name_to_var_map

    def print_name(self, obj, name_dict):
        return name_dict.get(obj._index)  # default is None

    def linearct_print_name(self, ct):
        return self.print_name(ct, self._linct_name_map)

    def logicalct_print_name(self, indicator):
        return self.print_name(indicator, self._lc_name_map)

    def qc_print_name(self, quad_constraint):
        return self.print_name(quad_constraint, self._qc_name_map)

    @staticmethod
    def _make_prefix_name(mobj, prefix, local_index, offset=1):
        prefixed_name = "{0:s}{1:d}".format(prefix, local_index + offset)
        return prefixed_name

    from docplex.mp.compat23 import mktrans

    __raw = " -+/\\<>"
    __cooked = "_mp____"

    _str_translate_table = mktrans(__raw, __cooked)
    _unicode_translate_table = {}
    for c in range(len(__raw)):
        _unicode_translate_table[ord(__raw[c])] = ord(__cooked[c])

    @staticmethod
    def _translate_chars2(raw_name):
        # noinspection PyUnresolvedReferences
        if isinstance(raw_name, unicode):
            char_mapping = TextModelPrinter._unicode_translate_table
        else:
            char_mapping = TextModelPrinter._str_translate_table
        return raw_name.translate(char_mapping)


    @staticmethod
    def _translate_chars3(raw_name):
        return raw_name.translate(TextModelPrinter._unicode_translate_table)

    def fix_name(self, mobj, prefix, local_index, hide_names):
        # INTERNAL
        raw_name = mobj.name
        if hide_names or mobj.is_generated() or not raw_name:
            return self._make_prefix_name(mobj, prefix, local_index, offset=1)
        else:
            return self._translate_chars(raw_name)

    @staticmethod
    def _generate_linear_obj_coefs(model, linear_obj_expr):
        # INTERNAL: print all variables and their coef in the linear part of the objective
        for v in model.iter_variables():
            yield v, linear_obj_expr.unchecked_get_coef(v)

    @staticmethod
    def _generate_linear_obj_coefs_smart(model, linear_obj_expr, selected_variables):
        # INTERNAL: used to print unreferenced variables in sos and pwl constraints and their coef in the linear part
        #  of the objective, in addition to variables in the objective
        for v in selected_variables:
            yield v, linear_obj_expr.unchecked_get_coef(v)

    def _print_lexpr(self, wrapper, num_printer, var_name_map, expr, print_constant=False, allow_empty=False,
                     force_first_plus=False):
        # prints an expr to a stream
        term_iter = expr.iter_sorted_terms()
        k = expr.get_constant() if print_constant else None
        self._print_expr_iter(wrapper, num_printer, var_name_map, term_iter,
                              constant=k, allow_empty=allow_empty,
                              force_first_plus=force_first_plus)

    def _print_expr_iter(self, wrapper, num_printer, var_name_map,
                         expr_iter,
                         allow_empty=False,
                         force_first_plus=False,
                         constant=None,
                         accept_zero=False):
        num2string_fn = num_printer.to_string
        c = 0
        if self.sort_variable_names:
            def varname(x):  # use function, since auto tuple unpacking not supported in py3
                v, _ = x
                return var_name_map[v._index]
            expr_iter = sorted(list(expr_iter), key=varname)
        for (v, coeff) in expr_iter:
            curr_token = ''
            if not accept_zero and not coeff:
                continue  # pragma: no cover

            if coeff < 0:
                curr_token += '-'
                wrote_sign = True
                coeff = - coeff
            elif c > 0 or force_first_plus:
                # here coeff is positive, we write the '+' only if term is non-first
                curr_token += '+'
                wrote_sign = True
            else:
                wrote_sign = False

            if 1 != coeff:
                if wrote_sign:
                    curr_token += ' '
                curr_token += num2string_fn(coeff)
            if wrote_sign or 1 != coeff:
                curr_token += ' '
            curr_token += var_name_map[v._index]

            wrapper.write(curr_token)
            c += 1

        printed_k = True
        if constant is not None:
            # here constant is a number
            if constant:
                if constant > 0:
                    if c > 0 or force_first_plus:
                        wrapper.write('+')
                wrapper.write(num2string_fn(constant))
            elif 0 == c and not allow_empty:
                wrapper.write('0')
            else:
                printed_k = False

        else:
            # constant is none here
            if not c and not allow_empty:
                # expr is empty, if we must print something, print 0
                wrapper.write('0')
            else:
                printed_k = False
        return c or printed_k

    def _print_qexpr_obj(self, wrapper, num_printer, var_name_map, quad_expr, force_initial_plus, use_double=True):
        # writes a quadratic expression
        # in the form [ 2a_ij a_i.a_j ] / 2
        # Note that all coefficients must be doubled due to the tQXQ formulation

        if force_initial_plus:
            wrapper.write('+')

        return self._print_qexpr_iter(wrapper, num_printer, var_name_map, quad_expr.iter_sorted_quads(),
                                      use_double=use_double)

    def _print_qexpr_iter(self, wrapper, num_printer, var_name_map, iter_quads, use_double=False):
        q = 0
        varname_getter = self._var_print_name
        for qvp, qk in iter_quads:
            curr_token = ''
            if 0 == qk:
                continue  # pragma: no cover
            if q == 0:
                wrapper.write('[')  # only once
            abs_qk = qk
            if qk < 0:
                curr_token += '-'
                abs_qk = - qk
                wrote_sign = True
            elif q > 0:
                curr_token += '+'
                wrote_sign = True
            else:
                wrote_sign = False
            if wrote_sign:
                curr_token += ' '

            # all coefficients must be doubled because of the []/2 pattern.
            abs_qk2 = 2 * abs_qk if use_double else abs_qk
            if abs_qk2 != 1:
                curr_token += num_printer.to_string(abs_qk2)
                curr_token += ' '

            if qvp.is_square():
                qv_name = varname_getter(qvp[0])
                curr_token += "%s^2" % qv_name
            else:
                qv1 = qvp[0]
                qv2 = qvp[1]
                curr_token += "%s*%s" % (varname_getter(qv1), varname_getter(qv2))

            wrapper.write(curr_token)

            q += 1

        if q:
            closer = ']/2' if use_double else ']'
            wrapper.write(closer)
        return q


class _ExportWrapper(object):
    """
    INTERNAL.
    """
    __new_line_sep = '\n'

    def __init__(self, oss, indent_str, line_width=80):
        self._oss = oss
        self._indent_str = indent_str
        self._line_width = line_width
        self._curr_line = ''
        self._wrote = False

    def wrote(self):  # pragma: no cover
        return not self._wrote

    def set_indent(self, new_indent):
        self._indent_str = new_indent

    def begin_line(self, indented=False):
        # reset dynamic line data
        self._wrote = False
        self._curr_line = self._indent_str if indented else ''

    # The 'write' function is invoked intensively when exporting a model.
    # Any piece of code that can be saved here will improve performance in a visible way.
    def write(self, token, separator=True):
        if token is not None:
            if len(self._curr_line) + len(token) >= self._line_width:
                # faster to write concatenated string, slightly faster to use '\n' instead of ref to static value
                self._oss.write(self._curr_line + '\n')
                #oss.write('\n')
                self._curr_line = self._indent_str + token
            else:
                # print one separator --before-- the token to be printed
                if separator and self._wrote:
                    self._curr_line += (' ' + token)
                else:
                    self._curr_line += token
            self._wrote = True

    def flush(self, print_newline=True, restart_from_empty_line=False):
        self._oss.write(self._curr_line)
        if print_newline:
            self._oss.write('\n')
        # Reset '_wrote' flag so that no separator will be added when writing first token of next line
        self._wrote = False
        # if reset, start a new line.
        self._curr_line = '' if restart_from_empty_line else self._indent_str

    def newline(self):
        self._oss.write('\n')



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2016, 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
This module allows to solve a model expressed as a CPO file using
a local CP Optimizer Interactive (cpoptimizer(.exe)).
"""

from docplex.cp.solution import *
from docplex.cp.utils import CpoException
import docplex.cp.solver.solver as solver

import subprocess
import sys
import time
import threading
import json
import signal, os


###############################################################################
##  Private constants
###############################################################################

# List of command ids that can be sent to solver
CMD_EXIT            = "Exit"           # End process (no data)
CMD_SET_CPO_MODEL   = "SetCpoModel"    # CPO model as string
CMD_SOLVE_MODEL     = "SolveModel"     # Complete solve of the model (no data)
CMD_START_SEARCH    = "StartSearch"    # Start search (no data)
CMD_SEARCH_NEXT     = "SearchNext"     # Get next solution (no data)
CMD_END_SEARCH      = "EndSearch"      # End search (no data)
CMD_REFINE_CONFLICT = "RefineConflict" # Refine conflict (no data)
CMD_PROPAGATE       = "Propagate"      # Propagate (no data)
CMD_RUN_SEEDS       = "RunSeeds"       # Run with multiple seeds.
CMD_ADD_CALLBACK    = "AddCallback"    # Add callback proxy to the solver

# List of events received from solver
EVT_VERSION_INFO        = "VersionInfo"        # Local solver version info (String in JSON format)
EVT_SUCCESS             = "Success"            # Success in last command execution
EVT_ERROR               = "Error"              # Error (data is error string)
EVT_TRACE               = "DebugTrace"         # Debugging trace
EVT_SOLVER_OUT_STREAM   = "OutStream"          # Solver output stream
EVT_SOLVER_WARN_STREAM  = "WarningStream"      # Solver warning stream
EVT_SOLVER_ERR_STREAM   = "ErrorStream"        # Solver error stream
EVT_SOLVE_RESULT        = "SolveResult"        # Solver result in JSON format
EVT_CONFLICT_RESULT     = "ConflictResult"     # Conflict refiner result in JSON format
EVT_CONFLICT_RESULT_CPO = "ConflictResultCpo"  # Conflict refiner result in CPO format
EVT_PROPAGATE_RESULT    = "PropagateResult"    # Propagate result in JSON format
EVT_RUN_SEEDS_RESULT    = "RunSeedsResult"     # Run seeds result (no data, all is in log)
EVT_CALLBACK_EVENT      = "CallbackEvent"      # Callback event. Data is event name.
EVT_CALLBACK_DATA       = "CallbackData"       # Callback data, following event. Data is JSON document.

# Max possible received data size in one message
_MAX_RECEIVED_DATA_SIZE = 1000000

# Python 2 indicator
IS_PYTHON_2 = (sys.version_info[0] == 2)

# Version of this client
CLIENT_VERSION = 3


###############################################################################
##  Public classes
###############################################################################

class LocalSolverException(CpoException):
    """ The base class for exceptions raised by the local solver client
    """
    def __init__(self, msg):
        """ Create a new exception
        Args:
            msg: Error message
        """
        super(LocalSolverException, self).__init__(msg)


class CpoSolverLocal(solver.CpoSolverAgent):
    """ Interface to a local solver through an external process """

    def __init__(self, solver, params, context):
        """ Create a new solver that solves locally with CP Optimizer Interactive.

        Args:
            solver:  Parent solver
            params:  Solving parameters
            context: Solver context
        Raises:
            CpoException if proxy executable does not exists
        """
        # Call super
        self.process = None
        self.active = True
        super(CpoSolverLocal, self).__init__(solver, params, context)

        # Check if executable file exists
        if context.execfile is None:
            raise CpoException("Executable file should be given in 'execfile' context attribute.")
        if not is_string(context.execfile):
            raise CpoException("Executable file should be given in 'execfile' as a string.")
        #if not os.path.isfile(context.execfile):
        #    raise CpoException("Executable file '" + str(context.execfile) + "' does not exists")

        # Create solving process
        cmd = [context.execfile]
        if context.parameters is not None:
            cmd.extend(context.parameters)
        context.log(2, "Solver exec command: '", ' '.join(cmd), "'")
        try:
            self.process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stderr=subprocess.STDOUT, stdout=subprocess.PIPE, universal_newlines=False)
        except:
            raise CpoException("Can not execute command '{}'. Please check availability of required executable file.".format(' '.join(cmd)))
        self.pout = self.process.stdin
        self.pin = self.process.stdout

        # Read initial version info from process
        self.version_info = None
        timer = threading.Timer(1, lambda: self.process.kill() if self.version_info is None else None)
        timer.start()
        evt, data = self._read_message()
        timer.cancel()
        if evt != EVT_VERSION_INFO:
            raise LocalSolverException("Unexpected event {} received instead of version info event {}.".format(evt, EVT_VERSION_INFO))
        self.version_info = verinf = json.loads(data)
        self.available_command = self.version_info['AvailableCommands']
        # Normalize information
        verinf['AgentModule'] = __name__

        context.log(3, "Local solver info: '", verinf, "'")

        # Check solver version if any
        sver = self.version_info.get('SolverVersion')
        mver = solver.get_model_format_version()
        if sver and mver and compare_natural(mver, sver) > 0:
            raise LocalSolverException("Solver version {} is lower than model format version {}.".format(sver, mver))

        # Send CPO model to process
        cpostr = self._get_cpo_model_string()
        self._write_message(CMD_SET_CPO_MODEL, cpostr)
        self._wait_json_result(EVT_SUCCESS)  # JSON stored
        context.log(3, "Model sent.")

        # Initialize CPO callback setting
        self.callback_added = False


    def __del__(self):
        # End solve
        self.end()


    def solve(self):
        """ Solve the model

        According to the value of the context parameter 'verbose', the following information is logged
        if the log output is set:
         * 1: Total time spent to solve the model
         * 2: The process exec file
         * 3: Content of the JSON response
         * 4: Solver traces (if any)
         * 5: Messages sent/receive to/from process

        Returns:
            Model solve result,
            object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        """
        # Add callback if needed
        self._add_callback_if_needed()

        # Start solve
        self._write_message(CMD_SOLVE_MODEL)

        # Wait JSON result
        jsol = self._wait_json_result(EVT_SOLVE_RESULT)

        # Build result object
        return self._create_result_object(CpoSolveResult, jsol)


    def start_search(self):
        """ Start a new search. Solutions are retrieved using method search_next().
        """
        # Add callback if needed
        self._add_callback_if_needed()

        self._write_message(CMD_START_SEARCH)


    def search_next(self):
        """ Get the next available solution.

        (This method starts search automatically.)

        Returns:
            Next model result (type CpoSolveResult)
        """

        # Request next solution
        self._write_message(CMD_SEARCH_NEXT)

        # Wait JSON result
        jsol = self._wait_json_result(EVT_SOLVE_RESULT)

        # Build result object
        return self._create_result_object(CpoSolveResult, jsol)


    def end_search(self):
        """ End current search.

        Returns:
            Last (fail) solve result with last solve information (type CpoSolveResult)
        """

        # Request end search
        self._write_message(CMD_END_SEARCH)

        # Wait JSON result
        jsol = self._wait_json_result(EVT_SOLVE_RESULT)

        # Build result object
        return self._create_result_object(CpoSolveResult, jsol)


    def abort_search(self):
        """ Abort current search.
        This method is designed to be called by a different thread than the one currently solving.
        """
        self.end()


    def refine_conflict(self):
        """ This method identifies a minimal conflict for the infeasibility of the current model.

        See documentation of :meth:`~docplex.cp.solver.solver.CpoSolver.refine_conflict` for details.

        Returns:
            Conflict result,
            object of class :class:`~docplex.cp.solution.CpoRefineConflictResult`.
        """
        # Ensure cpo model is generated with all constraints named
        if not self.context.model.name_all_constraints:
            self.context.model.name_all_constraints = True
            cpostr = self._get_cpo_model_string()
            self.context.model.name_all_constraints = False
            # Send CPO model to process
            self._write_message(CMD_SET_CPO_MODEL, cpostr)
            self._wait_event(EVT_SUCCESS)

        # Add callback if needed
        self._add_callback_if_needed()
        
        # Check if cpo format required
        pver = self.version_info.get('ProxyVersion')
        if self.context.add_conflict_as_cpo and pver and (int(pver) >= 9):
            # Request refine conflict with CPO format
            self._write_message(CMD_REFINE_CONFLICT, bytearray([1]))
            # Wait JSON result
            jsol = self._wait_json_result(EVT_CONFLICT_RESULT)
            # Wait for CPO conflict
            cposol = self._wait_event(EVT_CONFLICT_RESULT_CPO)
        else:
            # Request refine conflict
            self._write_message(CMD_REFINE_CONFLICT)
            # Wait JSON result
            jsol = self._wait_json_result(EVT_CONFLICT_RESULT)
            # No CPO conflict
            cposol = None

        # Build result object
        result = self._create_result_object(CpoRefineConflictResult, jsol)
        result.cpo_conflict = cposol
        return result


    def propagate(self):
        """ This method invokes the propagation on the current model.

        See documentation of :meth:`~docplex.cp.solver.solver.CpoSolver.propagate` for details.

        Returns:
            Propagation result,
            object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        """
        # Add callback if needed
        self._add_callback_if_needed()

        # Request propagation
        self._write_message(CMD_PROPAGATE)

        # Wait JSON result
        jsol = self._wait_json_result(EVT_PROPAGATE_RESULT)

        # Build result object
        return self._create_result_object(CpoSolveResult, jsol)


    def run_seeds(self, nbrun):
        """ This method runs *nbrun* times the CP optimizer search with different random seeds
        and computes statistics from the result of these runs.

        This method does not return anything. Result statistics are displayed on the log output
        that should be activated.

        Each run of the solver is stopped according to single solve conditions (TimeLimit for example).
        Total run time is then expected to take *nbruns* times the duration of a single run.

        Args:
            nbrun: Number of runs with different seeds.
        Returns:
            Run result, object of class :class:`~docplex.cp.solution.CpoRunResult`.
        Raises:
            CpoNotSupportedException: method not available in local solver.
        """
        # Add callback if needed
        self._add_callback_if_needed()

        # Check command availability
        if CMD_RUN_SEEDS not in self.available_command:
            raise CpoNotSupportedException("Method 'run_seeds' is not available in local solver '{}'".format(self.context.execfile))

        # Request run seeds
        nbfrm = bytearray(4)
        nbfrm[0] = (nbrun >> 24) & 0xFF
        nbfrm[1] = (nbrun >> 16) & 0xFF
        nbfrm[2] = (nbrun >> 8)  & 0xFF
        nbfrm[3] = nbrun         & 0xFF

        self._write_message(CMD_RUN_SEEDS, data=nbfrm)

        # Wait result
        self._wait_event(EVT_RUN_SEEDS_RESULT)

        # Build result object
        return self._create_result_object(CpoRunResult)


    def end(self):
        """ End solver and release all resources.
        """
        if self.active:
            self.active = False
            try:
                self._write_message(CMD_EXIT)
            except:
                pass
            try:
                self.pout.close()
            except:
                pass
            try:
                self.pin.close()
            except:
                pass
            try:
                self.process.kill()
            except:
                pass
            try:
                self.process.wait()
            except:
                pass
            self.process = None
            super(CpoSolverLocal, self).end()


    def _wait_event(self, xevt):
        """ Wait for a particular event while forwarding logs if any.
        Args:
            xevt: Expected event
        Returns:
            Message data
        Raises:
            LocalSolverException if an error occurs
        """
        # Initialize first error string to enrich exception if any
        firsterror = None

        # Read events
        while True:
            # Read and process next message
            evt, data = self._read_message()

            if evt == xevt:
                return data

            elif evt in (EVT_SOLVER_OUT_STREAM, EVT_SOLVER_WARN_STREAM):
                if data:
                    # Store log if required
                    if self.log_enabled:
                        self._add_log_data(data)

            elif evt == EVT_SOLVER_ERR_STREAM:
                if data:
                    if firsterror is None:
                        firsterror = data.replace('\n', '')
                    out = self.log_output if self.log_output is not None else sys.stdout
                    out.write("ERROR: {}\n".format(data))
                    out.flush()

            elif evt == EVT_TRACE:
                self.context.log(4, "ANGEL: " + data)

            elif evt == EVT_ERROR:
                if firsterror is not None:
                    data += " (" + firsterror + ")"
                self.end()
                raise LocalSolverException("Solver error: " + data)

            elif evt == EVT_CALLBACK_EVENT:
                event = data
                # Read data
                evt, data = self._read_message()
                assert evt == EVT_CALLBACK_DATA
                res = self._create_result_object(CpoSolveResult, data)
                self.solver._notify_callback_event(event, res)

            else:
                self.end()
                raise LocalSolverException("Unknown event received from local solver: " + str(evt))


    def _wait_json_result(self, evt):
        """ Wait for a JSON result while forwarding logs if any.
        Args:
            evt: Event to wait for
        Returns:
            JSON solution string, decoded from UTF8
        """

        # Wait JSON result
        data = self._wait_event(evt)

        # Store last json result
        self._set_last_json_result_string(data)
        self.context.log(3, "JSON result:\n", data)

        return self.last_json_result


    def _write_message(self, cid, data=None):
        """ Write a message to the solver process
        Args:
            cid:   Command name
            data:  Data to write, already encoded in UTF8 if required
        """
        # Encode elements
        stime = time.time()
        cid = cid.encode('utf-8')
        if is_string(data):
            data = data.encode('utf-8')
        nstime = time.time()
        self.process_infos.incr(CpoProcessInfos.TOTAL_UTF8_ENCODE_TIME, nstime - stime)

        # Build header
        tlen = len(cid)
        if data is not None:
            tlen += len(data) + 1
        if tlen > 0xffffffff:
            raise LocalSolverException("Try to send a message with length {}, greater than {}.".format(tlen, 0xffffffff))
        frame = bytearray(6)
        frame[0] = 0xCA
        frame[1] = 0xFE
        frame[2] = (tlen >> 24) & 0xFF
        frame[3] = (tlen >> 16) & 0xFF
        frame[4] = (tlen >> 8)  & 0xFF
        frame[5] = tlen         & 0xFF

        # Add data if any
        self.context.log(5, "Send message: cmd=", cid, ", tsize=", tlen)
        if data is None:
            frame = frame + cid
        else:
            frame = frame + cid + bytearray(1) + data

        # Write message frame
        self.pout.write(frame)
        self.pout.flush()

        # Update statistics
        self.process_infos.incr(CpoProcessInfos.TOTAL_DATA_SEND_TIME, time.time() - nstime)
        self.process_infos.incr(CpoProcessInfos.TOTAL_DATA_SEND_SIZE, len(frame))


    def _read_message(self):
        """ Read a message from the solver process
        Returns:
            Tuple (evt, data)
        """
        # Read message header
        frame = self._read_frame(6)
        if (frame[0] != 0xCA) or (frame[1] != 0xFE):
            erline = frame + self._read_error_message()
            erline = erline.decode()
            self.end()
            raise LocalSolverException("Invalid message header. Possible error generated by solver: " + erline)

        # Read message data
        tsize = (frame[2] << 24) | (frame[3] << 16) | (frame[4] << 8) | frame[5]
        data = self._read_frame(tsize)

        # Split name and data
        ename = 0
        while (ename < tsize) and (data[ename] != 0):
            ename += 1

        # Decode name and data
        stime = time.time()
        if ename == tsize:
            # Command only, no data
            evt = data.decode('utf-8')
            data = None
        else:
            # Split command and data
            evt = data[0:ename].decode('utf-8')
            data = data[ename+1:].decode('utf-8')

        # Update statistics
        self.process_infos.incr(CpoProcessInfos.TOTAL_UTF8_DECODE_TIME, time.time() - stime)
        self.process_infos.incr(CpoProcessInfos.TOTAL_DATA_RECEIVE_SIZE, tsize + 6)

        # Log received message
        self.context.log(5, "Read message: ", evt, ", data: '", data, "'")

        return evt, data


    def _read_frame(self, nbb):
        """ Read a byte frame from input stream
        Args:
            nbb:  Number of bytes to read
        Returns:
            Byte array
        """
        # Read data
        data = self.pin.read(nbb)
        if len(data) != nbb:
            if len(data) == 0:
                # Check if first read of data
                if self.process_infos.get(CpoProcessInfos.TOTAL_DATA_RECEIVE_SIZE, 0) == 0:
                    if IS_WINDOWS:
                        raise LocalSolverException("Nothing to read from local solver process. Possibly not started because cplex dll is not accessible.")
                    else:
                        raise LocalSolverException("Nothing to read from local solver process. Check its availability.")
                else:
                    try:
                        self.process.wait()
                        rc = self.process.returncode
                    except:
                        rc = "unknown"
                    raise LocalSolverException("Nothing to read from local solver process. Process seems to have been stopped (rc={}).".format(rc))
            else:
                raise LocalSolverException("Read only {} bytes when {} was expected.".format(len(data), nbb))

        # Return
        if IS_PYTHON_2:
            data = bytearray(data)
        return data


    def _read_error_message(self):
        """ Read stream to search for error line end. Called when wrong input is detected,
        to try to read an "Assertion failed" message for example.
        Returns:
            Byte array
        """
        data = []
        bv = self.pin.read(1)
        if IS_PYTHON_2:
            while (bv != '') and (bv != '\n'):
                data.append(ord(bv))
                bv = self.pin.read(1)
                data = bytearray(data)
        else:
            while (bv != b'') and (bv != b'\n'):
                data.append(ord(bv))
                bv = self.pin.read(1)

        return bytearray(data)


    def _add_callback_if_needed(self):
        """ Ask solver to add callback if needed. """
        if not self.callback_added and self.solver.callbacks:
            # Check solver version
            aver = self.version_info.get('AngelVersion', 0)
            sver = self.version_info.get('SolverVersion', "1")
            if (compare_natural(sver, "12.10") >= 0) and (aver >= 8):
                self._write_message(CMD_ADD_CALLBACK)
                self._wait_event(EVT_SUCCESS)
                self.callback_added = True
                self.context.log(3, "CPO callback created.")
            else:
                raise LocalSolverException("This version of the CPO solver does not support solver callbacks. Solver: {}, agent: {}".format(sver, aver))


###############################################################################
##  Public functions
###############################################################################

from docplex.cp.model import CpoModel

def get_solver_info():
    """ Get the information data of the local CP solver that is target by the solver configuration.

    This method creates a CP solver to retrieve this information, and end it immediately.
    It returns a dictionary with various information, as in the following example:
    ::
    {
       "AngelVersion" : 5,
       "SourceDate" : "Sep 12 2017",
       "SolverVersion" : "12.8.0.0",
       "IntMin" : -9007199254740991,
       "IntMax" : 9007199254740991,
       "IntervalMin" : -4503599627370494,
       "IntervalMax" : 4503599627370494,
       "AvailableCommands" : ["Exit", "SetCpoModel", "SolveModel", "StartSearch", "SearchNext", "EndSearch", "RefineConflict", "Propagate", "RunSeeds"]
    }

    Returns:
        Solver information dictionary, or None if not available.
    """
    try:
        with solver.CpoSolver(CpoModel()) as slvr:
            if isinstance(slvr.agent, CpoSolverLocal):
                return slvr.agent.version_info
    except:
        pass
    return None







# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore

# import threading
#
# __before_dict_ready = threading.Event()
# __results_ready = threading.Event()
# __before_solve_dict = {}
# __after_solve_dict = {}
# __solve_status = None
# __has_solution = False
# __objective_value = None
# solve_hook = None



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

from collections import namedtuple, defaultdict
from docplex.mp.utils import is_string
from docplex.mp.constants import ComparisonType, VarBoundType
from docplex.mp.context import check_credentials
from docplex.mp.cloudutils import context_must_use_docloud

from docplex.mp.utils import str_maxed
from docplex.mp.publish import PublishResultAsDf

import warnings

TConflictConstraint = namedtuple("_TConflictConstraint", ["name", "element", "status"])


def trim_field(element):
    return str_maxed(element, maxlen=50)


def to_output_table(conflicts, use_df=True):
    # Returns the output tables, as df if pandas is available or as a list
    # of named tuple ['Type', 'Status', 'Name', 'Expression']
    columns = ['Type', 'Status', 'Name', 'Expression']
    TOutputTables = namedtuple('TOutputTables', columns)

    def convert_to_pandas_df(c):
        return {'Type': 'Constraint',
                'Status': c.status.name if c.status is not None else '',
                'Name': c.name or '',
                'Expression': trim_field(c.element)}

    def convert_to_namedtuples(c):
        return TOutputTables('Constraint',
                             c.status,
                             c.name or '',
                             trim_field(c.element))

    pandas = None
    if use_df:
        try:
            import pandas
        except ImportError:  # pragma: no cover
            print("pandas module not found...")
            pandas = None

    data_converter = convert_to_pandas_df if pandas and use_df else convert_to_namedtuples
    output_data = list(map(data_converter, conflicts))

    if use_df:
        return pandas.DataFrame(columns=columns, data=output_data)
    else:
        return output_data


class ConflictRefinerResult(object):
    """ This class contains all conflicts as returned by the conflict refiner.

    A conflict refiner result contains a list of named tuples of type ``TConflictConstraint``,
    the fields of which are:

        - an enumerated value of type  ``docplex.mp.constants.ConflictStatus`` that indicates the
                  conflict status type (Excluded, Possible_member, Member...).
        - the name of the constraint or None if the constraint corresponds to a variable lower or upper bound.
        - a modeling object involved in the conflict:
            can be either a constraint or a wrapper representing a variable upper or lower bound.


        *New in version 2.11*
    """

    def __init__(self, conflicts, refined_by=None):
        self._conflicts = conflicts
        assert refined_by is None or is_string(refined_by)
        self._refined_by = refined_by

    @property
    def refined_by(self):
        '''
        Returns a string indicating how the conflicts were produced.

        - If the conflicts are created by a program, this field returns None.
        - If the conflicts originated from a local CPLEX run, this method returns 'cplex_local'.
        - If the conflicts originated from a DOcplexcloud run, this method returns 'cplex_cloud'.

        Returns:
            A string, or None.

        '''
        return self._refined_by

    def __iter__(self):
        return self.iter_conflicts()

    def __len__(self):
        """ Redefintion of maguic method __len__.

        Allows calling len() on an instance of ConflictRefinerResult
        to get the number of conflicts

        :return: the number of conflicts.
        """
        return len(self._conflicts)

    def iter_conflicts(self):
        """ Returns an iterator on conflicts (named tuples)

        :return: an iterator
        """
        return iter(self._conflicts)

    @property
    def number_of_conflicts(self):
        """ This property returns the number of conflicts. """
        return len(self._conflicts)

    def display(self):
        """ Displays all conflicts.

        """
        print('conflict(s): {0}'.format(self.number_of_conflicts))
        for conflict in self.iter_conflicts():
            st = conflict.status
            elt = conflict.element
            if hasattr(conflict.element, 'as_constraint'):
                ct = conflict.element.as_constraint()
                label = elt.short_typename
            else:
                ct = elt
                label = ct.__class__.__name__
            print("  - status: {1}, {0}: {2!s}".format(label, st.name, str_maxed(ct, maxlen=40)))

    def display_stats(self):
        """ Displays statistics on conflicts.

        Display show many conflict elements per type.
        """
        def elt_typename(elt):
            try:
                return elt.short_typename.lower()
            except AttributeError:  # pragma: no cover
                return elt.__class__.__name__.lower()

        ncf = self.number_of_conflicts
        print('conflict{1}: {0}'.format(ncf, "s" if ncf > 1 else ""))
        cf_stats = defaultdict(lambda: 0)
        for conflict in self.iter_conflicts():
            elt_type = elt_typename(conflict.element)
            cf_stats[elt_type] += 1
        for eltt, count in cf_stats.items():
            if count:
                print("  - {0}{2}: {1}".format(eltt, count, ("s" if count > 1 else "")))

    def as_output_table(self, use_df=True):
        return to_output_table(self, use_df)


    def print_information(self):
        """ Similar as `display_stats`
        """
        self.display_stats()


class VarBoundWrapper(object):
    # INTERNAL

    def __init__(self, dvar):
        self._var = dvar

    @property
    def var(self):
        return self._var

    @property
    def index(self):
        return self._var.index

    @classmethod
    def short_typename(cls):  # pragma: no cover
        return "Variable Bound"

    def as_constraint(self):  # pragma: no cover
        raise NotImplementedError

    def as_constraint_from_symbol(self, op_symbol):
        self_var = self.var
        var_lb = self.var.lb
        op = ComparisonType.cplex_ctsense_to_python_op(op_symbol)
        ct = op(self_var, var_lb)
        return ct

    @classmethod
    def make_wrapper(cls, var, bound_type):
        if bound_type == VarBoundType.LB:
            return VarLbConstraintWrapper(var)
        elif bound_type == VarBoundType.UB:
            return VarUbConstraintWrapper(var)
        else:
            return None


class VarLbConstraintWrapper(VarBoundWrapper):
    """
    This class is a wrapper for a model variable and its associated lower bound.

    Instances of this class are created by the ``refine_conflict`` method when the conflict involves
    a variable lower bound. Each of these instances is then referenced by a ``TConflictConstraint`` namedtuple
    in the conflict list returned by ``refine_conflict``.

    To check whether the lower bound of a variable causes a conflict, wrap the variable and
    include the resulting constraint in a ConstraintsGroup.
    """

    def __init__(self, dvar):
        super(VarLbConstraintWrapper, self).__init__(dvar)
        self._var = dvar

    @property
    def short_typename(self):
        return "Lower Bound"

    def as_constraint(self):
        return self.as_constraint_from_symbol('G')


class VarUbConstraintWrapper(VarBoundWrapper):
    """
    This class is a wrapper for a model variable and its associated upper bound.

    Instances of this class are created by the ``refine_conflict`` method when the conflict involves
    a variable upper bound. Each of these instances is then referenced by a ``TConflictConstraint`` namedtuple
    in the conflict list returned by ``refine_conflict``.

    To check whether the upper bound of a variable causes a conflict, wrap the variable and
    include the resulting constraint in a ConstraintsGroup.
    """

    def __init__(self, var):
        super(VarUbConstraintWrapper, self).__init__(var)

    @property
    def short_typename(self):
        return "Upper Bound"

    def as_constraint(self):
        return self.as_constraint_from_symbol('L')


class ConstraintsGroup(object):
    """
    This class is a container for the definition of a group of constraints.
    A preference for conflict refinement is associated to the group.

    Groups may be assigned preference. A group with a higher preference is more likely to be included in the conflict.
    A negative value specifies that the corresponding group should not be considered in the computation
    of a conflict. In other words, such groups are not considered part of the model. Groups with a preference of 0 (zero)
    are always considered to be part of the conflict.

    Args:
        preference: A floating-point number that specifies the preference for the group. The higher the number, the
                    higher the preference.
    """

    __slots__ = ('_preference', '_cts')

    def __init__(self, preference=1.0, cts=None):
        self._preference = preference
        self._cts = []
        if cts is not None:
            self.add_constraints(cts)

    @classmethod
    def from_var(cls, dvar, bound_type, pref):
        """ A class method to build a group fromone variable.

        :param dvar: The variable whose bound is part of the conflict.
        :param bound_type: An enumerated value of type `VarBoundType`
        :param pref: a numerical preference.

        :return: an instance of ConstraintsGroup.

        See Also:
            :class:`docplex.mp.constants.VarBoundType`
        """
        cgg = cls(preference=pref, cts=VarBoundWrapper.make_wrapper(dvar, bound_type))
        return cgg

    @property
    def preference(self):
        return self._preference

    def add_one(self, x):
        if x is not None:
            self._cts.append(x)

    def add_constraint(self, ct):
        self._cts.append(ct)

    def add_constraints(self, cts):
        try:
            for ct in cts:
                self.add_one(ct)
        except TypeError:  # not iterable.
            self.add_one(cts)

    def iter_constraints(self):
        return iter(self._cts)


class ConflictRefiner(PublishResultAsDf, object):
    ''' This class is an abstract algorithm; it operates on interfaces.

    A conflict is a set of mutually contradictory constraints and bounds within a model.
    Given an infeasible model, the conflict refiner can identify conflicting constraints and bounds
    within it. CPLEX refines an infeasible model by examining elements that can be removed from the
    conflict to arrive at a minimal conflict.
    '''

    # static variables for output
    output_table_property_name = 'conflicts_output'
    default_output_table_name = 'conflicts.csv'
    output_table_using_df = True

    def __init__(self, output_processing=None):
        self.output_table_customizer = output_processing

    @classmethod
    def _make_atomic_ct_groups(cls, mdl_iter, pref):
        # returns a list of singleton groups from a model iterator and a numerical preference.
        lcgrps = [ConstraintsGroup(pref, ct) for ct in mdl_iter]
        return lcgrps

    @classmethod
    def var_bounds(cls, mdl, pref=4.0, include_infinity_bounds=True):
        """ Returns a list of singleton groups with variable bounds.

        This method a list of ConstraintGroup objects, each of which contains a variabel bound.
        It replicate sthe behavior of the CPLEX interactive optimizer, that is, it returns

        - lower bounds for non-binary variables if different from 0
        - upper bound for non-binary-variables if non-default

        For binary variables, bounds are not considered, unless the variable is bound; more precisely:
            - lower bound is included if >= 0.5
            - upper bound is included if <= 0.5

        :param mdl: The model being analyzed for conflicts,
        :param pref: the preference for variable bounds, the defaut is 4.0
        :param include_infinity_bounds: a flag indicating whether infi

        :return: a list of `ConstraintsGroup` objects.
        """
        grps = []
        mdl_inf = mdl.infinity
        for dv in mdl.iter_variables():
            lb, ub = dv.lb, dv.ub
            if not dv.is_binary():
                if lb != 0:
                    if include_infinity_bounds or lb > - mdl_inf:
                        grps.append(ConstraintsGroup.from_var(dv, VarBoundType.LB, pref))
                if include_infinity_bounds or ub < mdl_inf:
                    grps.append(ConstraintsGroup.from_var(dv, VarBoundType.UB, pref))
            else:
                if lb >= 0.5:
                    grps.append(ConstraintsGroup.from_var(dv, VarBoundType.LB, pref))
                if ub <= 0.5:
                    grps.append(ConstraintsGroup.from_var(dv, VarBoundType.UB, pref))
        return grps

    @classmethod
    def linear_constraints(cls, mdl, pref=2.0):
        return cls._make_atomic_ct_groups(mdl.iter_linear_constraints(), pref)

    @classmethod
    def logical_constraints(cls, mdl, pref=1.0):
        return cls._make_atomic_ct_groups(mdl.iter_logical_constraints(), pref)

    @classmethod
    def quadratic_constraints(cls, mdl, pref=1.0):
        return cls._make_atomic_ct_groups(mdl.iter_quadratic_constraints(), pref)

    def refine_conflict(self, mdl, preferences=None, groups=None, display=False, **kwargs):
        """ Starts the conflict refiner on the model.

        Args:
            mdl: The model to be relaxed.
            preferences: A dictionary defining constraint preferences.
            groups: A list of ConstraintsGroups.
            display: a boolean flag (default is True); if True, displays the result at the end.
            kwargs: Accepts named arguments similar to solve.

        Returns:
            An object of type `ConflictRefinerResut` which holds all information about
            the minimal conflict.

        See Also:
            :class:`ConflictRefinerResult`

        """

        if mdl.has_multi_objective():
            mdl.fatal("Conflict refiner is not supported for multi-objective")

        # take into account local argument overrides
        context = mdl.prepare_actual_context(**kwargs)

        # log stuff
        saved_context_log_output = mdl.context.solver.log_output
        saved_log_output_stream = mdl.log_output

        try:
            mdl.set_log_output(context.solver.log_output)

            forced_docloud = context_must_use_docloud(context, **kwargs)

            results = None

            have_credentials = False
            if context.solver.docloud:
                have_credentials, error_message = check_credentials(context.solver.docloud)
                if error_message is not None:
                    warnings.warn(error_message, stacklevel=2)
            if forced_docloud:
                if have_credentials:
                    results = self._refine_conflict_cloud(mdl, context, preferences, groups)
                else:
                    mdl.fatal("DOcplexcloud context has no valid credentials: {0!s}", context.solver.docloud)
            # from now on docloud_context is None
            elif mdl.environment.has_cplex:
                # if CPLEX is installed go for it
                results = self._refine_conflict_local(mdl, context, preferences, groups)
            elif have_credentials:
                # no context passed as argument, no Cplex installed, try model's own context
                results = self._refine_conflict_cloud(mdl, context, preferences, groups)
            else:
                # no way to solve.. really
                return mdl.fatal("CPLEX DLL not found: please provide DOcplexcloud credentials")

            # write conflicts table.write_output_table() handles everything related to
            # whether the table should be published etc...
            if self.is_publishing_output_table(mdl.context):
                self.write_output_table(results.as_output_table(self.output_table_using_df), mdl.context)
            if display:
                results.display_stats()
            return results
        finally:
            if saved_log_output_stream != mdl.log_output:
                mdl.set_log_output_as_stream(saved_log_output_stream)
            if saved_context_log_output != mdl.context.solver.log_output:
                mdl.context.solver.log_output = saved_context_log_output

    # noinspection PyMethodMayBeStatic
    def _refine_conflict_cloud(self, mdl, context, preferences=None, groups=None):
        # INTERNAL
        docloud_context = context.solver.docloud
        parameters = context.cplex_parameters
        # see if we can reuse the local docloud engine if any?
        docloud_engine = mdl._engine_factory.new_docloud_engine(model=mdl,
                                                                docloud_context=docloud_context,
                                                                log_output=context.solver.log_output_as_stream)

        mdl.notify_start_solve()
        mdl._fire_start_solve_listeners()
        conflict = docloud_engine.refine_conflict(mdl, preferences=preferences, groups=groups, parameters=parameters)

        mdl._fire_end_solve_listeners(conflict is not None, None)
        #
        return conflict

    # noinspection PyMethodMayBeStatic
    def _refine_conflict_local(self, mdl, context, preferences=None, groups=None):
        parameters = context.cplex_parameters
        self_engine = mdl.get_engine()
        return self_engine.refine_conflict(mdl, preferences, groups, parameters)

    @staticmethod
    def display_conflicts(conflicts):
        """
        This method displays a formatted representation of the conflicts that are provided.

        Args:
           conflicts: An instance of ``ConflictRefinerResult``
        """
        warnings.warn("deprecated: use ConflictRefinerresult.display", DeprecationWarning)
        conflicts.display()
